with Messages,Ada.Text_IO,Reps,Reps.IO;
use Messages;

package body Doubling is

   package Txt_Int_IO is new Ada.Text_IO.Integer_IO(Integer);

   function Alpha(Y,X,Eps: in Scalar) return Scalar is
   begin
      return Scal(One)/(Two*Sqrt(Y-X/(Scal(One)+Eps))-Two*Sqrt(Y-Scal(One)));
   end Alpha;

   function Beta(Y,X,Eps: in Scalar) return Scalar is
   begin
      return Sqrt(Y-X/(Scal(One)+Eps));
   end Beta;

   function Gamma(Y: in Scalar) return Scalar is
   begin
      return Sqrt(Y-Scal(One));
   end Gamma;

   function LL(Y,X,Eps: in Scalar) return Scalar is
   begin
      return -R1/Alpha(Y,X,Eps)-Beta(Y,X,Eps);
   end LL;

   function RR(Y,X,Eps: in Scalar) return Scalar is
   begin
      return -L1/Alpha(Y,X,Eps)-Beta(Y,X,Eps);
   end RR;

   function T(Y,X,Eps,Z: in Scalar) return Scalar is
   begin
      return -Alpha(Y,X,Eps)*(Z+Beta(Y,X,Eps));
   end T;

   function TI(Y,X,Eps,Z: in Scalar) return Scalar is
   begin
      return -Z/Alpha(Y,X,Eps)-Beta(Y,X,Eps);
   end TI;

   function RI2(T,S: in Scalar) return Scalar is
   begin
      return  P2*(T-Te)*(R1+Scal(Half))+
        Scal(Zero)*(S-Es)*R1+R2;
   end RI2;

   function LI2(T,S: in Scalar) return Scalar is
   begin
      return  P2*(T-Te)*(L1+Scal(Half))+
       Scal(Zero)*(S-Es)*L1+L2;
   end LI2;

   function RI3(T,S: in Scalar) return Scalar is
   begin
      return  P3*(T-Te)*(R1+Scal(Half))+
        Scal(Zero)*(S-Es)*R1+R3;
   end RI3;

   function LI3(T,S: in Scalar) return Scalar is
   begin
      return  P3*(T-Te)*(L1+Scal(Half))+
        Scal(Zero)*(S-Es)*L1+L3;
   end LI3;

   ---------------------------------------------

   function Zeta(Z: in Scalar) return Scalar is
   begin
      return ( Sqrt(Z+Scal(One)) - Sqrt(Scal(One)-Z) )/(Sqrt(Z+Scal(One))+Sqrt(Scal(One)-Z));
   end Zeta;

   function ZetaI(Z: in Scalar) return Scalar is
   begin
      return Two*Z/(Scal(One)+Sqr(Z));
   end ZetaI;


   procedure ZetaI(Z1,Z2: in out Scalar) is
      Z: Scalar;
   begin
      --Two*(Z1+I*Z2)*(1+Z1^2-Z2^2+2*I*Z1*Z2)
      -- /((1+Z1^2-Z2^2)^2+4*(Z1*Z2)^2)
      Z:=Two*(Z1*( Scal(One)+Sqr(Z1)-Sqr(Z2) )+
         Two*Sqr(Z2)*Z1)/( Sqr(Scal(One)+(Sqr(Z1)-Sqr(Z2)) )+Scal(4.0)*Sqr(Z1*Z2) );
      Z2:=Two*(Z2*(Scal(One)+Sqr(Z1)-Sqr(Z2))-Two*Sqr(Z1)*Z2)/( Sqr(Scal(One)+(Sqr(Z1)-Sqr(Z2)) )+Scal(4.0)*Sqr(Z1*Z2) );
      Z1:=Z;
   end ZetaI;

   function Sigma1(Z: in Scalar) return Scalar is
   begin
      return (Exp(Kappa1*Log(Z+Scal(One)))-Exp(Kappa1*Log(Scal(One)-Z)))  / (Exp(Kappa1*Log(Z+Scal(One)))+Exp(Kappa1*Log(Scal(One)-Z))) ;
   end Sigma1;

   function Sigma1I(Z: in Scalar) return Scalar is
   begin
      return (Exp(Log(Z+Scal(One))/Kappa1)-Exp(Log(Scal(One)-Z)/Kappa1) )  / (Exp(Log(Z+Scal(One))/Kappa1)+Exp(Log(Scal(One)-Z)/Kappa1)) ;
   end Sigma1I;

   procedure Sigma1I(Z1,Z2: in out Scalar)  is
      S1,S2,S,T1,T2,T: Scalar;
   begin
      S1:=(Z1+Scal(One));
      S2:=Z2;
      Ln(S1,S2);
      S1:=S1/Kappa1;
      S2:=S2/Kappa1;
      S:= Exp(S1)*Cos(S2);
      S2:=Exp(S1)*Sin(S2);
      S1:=S;
      -------------------------------------------
      T1:=(Scal(One)-Z1);
      T2:=-Z2;
      Ln(T1,T2);
      T1:=T1/Kappa1;
      T2:=T2/Kappa1;
      T:= Exp(T1)*Cos(T2);
      T2:=Exp(T1)*Sin(T2);
      T1:=T;
      -------------------------------------------
      --( (S1-T1)+I*(S2-T2))*((S1+T1)-I(S2+T2))    / ( Sqr(S1+T1)+Sqr(S2+T2)  );

      Z1:=( (S1-T1)*(S1+T1) + (S2-T2)*(S2+T2) )/( Sqr(S1+T1)+Sqr(S2+T2));
      Z2:=Two*(S2*T1-S1*T2)/( Sqr(S1+T1)+Sqr(S2+T2) );
   end Sigma1I;

   function Q1(Z: in Scalar) return Scalar is
   begin
      return Half*(R1-L1)*Z+Half*(R1+L1);
   end Q1;

   function Q1I(Z: in Scalar) return Scalar is
   begin
      return (Two*Z-(R1+L1))/(R1-L1);
   end Q1I;

   procedure Q1I(Z1,Z2: in out Scalar) is
   begin
      Z1:=(Two*Z1-(R1+L1))/(R1-L1);
      Z2:=(Two*Z2)/(R1-L1);
   end Q1I;

   function M1(Z: in Scalar) return Scalar is
   Ak: constant Scalar:=Sigma1I(Q1I(Scal(-Half)));
   begin
      return (Z+Ak)/(Scal(One)-Ak*Z);
   end M1;

   function M1I(Z: in Scalar) return Scalar is
      Ak: constant Scalar:=Sigma1I(Q1I(Scal(-Half)));
   begin
      return (Z-Ak)/(Scal(One)+Ak*Z);
   end M1I;

   procedure M1I(Z1,Z2: in out Scalar) is
      Ak: constant Scalar:=Sigma1I(Q1I(Scal(-Half)));
      R: Scalar:=Sqr(Scal(One)+Ak*Z1) + Sqr(Ak*Z2);
   begin
      -- (Z1+I*Z2-Ak)/(Scal(One)+Ak*(Z1+I*Z2));
      -- (Z1+I*Z2-Ak)*(Scal(One)+Ak*Z1-Ak*I*Z2);
      Z1:= ( (Z1-Ak)*(Scal(One)+Ak*Z1)+Ak*Sqr(Z2)) / R;
      Z2:= ( Z2+Sqr(Ak)*Z2 ) / R;
   end M1I;

   function Theta1(Z: in Scalar) return Scalar is
   begin
      return Q1(Sigma1(M1(Zeta(Z)))) ;
   end Theta1;

   function Phi1(Z: in Scalar) return Scalar is
   begin
      return ZetaI(M1I(Sigma1I(Q1I(Z)))) ;
   end Phi1;

   procedure Phi1(Z1,Z2: in out Scalar)  is
      R: Scalar:=Half*(R1-L1);
      Cen1: Scalar:=Half*(R1+L1);
      Theta1: Scalar:=Half*(ScalPi-ScalPi*Kappa1);
      Y1: Scalar:=R*Sin(Theta1)/Cos(Theta1);
      R1: Scalar:=Sqrt( Sqr(R)+Sqr(Y1) );
      ZZ2: Scalar:=Z2;
      ZZ1: Scalar:=Z1;
   begin
      if Sup(Z2)>Zero and Sup(Sqr(Z1-Cen1)+Sqr(Z2+abs(Y1)))>Inf(Sqr(R1)) then
         New_Line;
         Put("ACHTUNG: Sqr of radius of the Poincare neighborhood=");Show(Sqr(R1));
         Put("ACHTUNG: Sqr of the distance to the center=");Show( Sqr(Z1-Cen1)+Sqr(Z2+abs(Y1))  );
        -- Message("Doubling.Phi1",Domain_Error);
      elsif Inf(Z2)<Zero and Sup(Sqr(Z1-Cen1)+Sqr(Z2-abs(Y1)))>Inf(Sqr(R1)) then
         New_Line;
         Put("ACHTUNG: Sqr of radius of the Poincare neighborhood=");Show(Sqr(R1));
         Put("ACHTUNG: Sqr of the distance to the center=");Show( Sqr(Z1-Cen1)+Sqr(Z2-abs(Y1))  );

        -- Message("Doubling.Phi1",Domain_Error);
      end if;


      --New_Line;
      Q1I(Z1,Z2);
      --Put("Z11=");Show(Z1);
      --Put("Z12=");Show(Z2);
      Sigma1I(Z1,Z2);
      --Put("Z21=");Show(Z1);
      --Put("Z22=");Show(Z2);
      M1I(Z1,Z2);
      --Put("Z31=");Show(Z1);
      --Put("Z32=");Show(Z2);
      ZetaI(Z1,Z2);
      --Put("Z41=");Show(Z1);
      --Put("Z42=");Show(Z2);
      if  Inf(ZZ2)>Zero and Sup(Z2)<Zero then
         New_Line;
         Put("AVIS: Initial Im(Z)=");Show(ZZ2);
         Put("AVIS: Im(Phi1(Z))=");Show(Z2);
      end if;

      if  Sup(ZZ2)<Zero and Inf(Z2)>Zero then
         New_Line;
         Put("AVIS: Initial Im(Z)=");Show(ZZ2);
         Put("AVIS: Im(Phi1(Z))=");Show(Z2);
      end if;
      end Phi1;

   ---------------------------------------------

   function Zeta2(Z,T,S: in Scalar) return Scalar is
   begin
      return ( Sqrt(Z-LI2(T,S)) - Sqrt(RI2(T,S)-Z) )/(Sqrt(Z-LI2(T,S))+Sqrt(RI2(T,S)-Z));
   end Zeta2;

   procedure Zeta2(Z1,Z2: in out Scalar; T,S: in Scalar) is
      S1,S2,SS,T1,T2,TT: Scalar;
   begin
      S1:=Z1-LI2(T,S);
      S2:=Z2;
      Ln(S1,S2);
      S1:=Half*S1;
      S2:=Half*S2;
      SS:=Exp(S1)*Cos(S2);
      S2:=Exp(S1)*Sin(S2);
      S1:=SS;

      ---------------------------------------------

      T1:=RI2(T,S)-Z1;
      T2:=-Z2;
      Ln(T1,T2);
      T1:=Half*T1;
      T2:=Half*T2;
      TT:=Exp(T1)*Cos(T2);
      T2:=Exp(T1)*Sin(T2);
      T1:=TT;
      ---------------------------------------------
      Z1:=( (S1-T1)*(S1+T1) + (S2-T2)*(S2+T2) )/( Sqr(S1+T1)+Sqr(S2+T2) );
      Z2:=Two*(S2*T1-S1*T2)/( Sqr(S1+T1)+Sqr(S2+T2) );
   end Zeta2;

   function Zeta2I(Z,T,S: in Scalar) return Scalar is
   begin
      return (Sqr(Scal(One)+Z)*RI2(T,S)+Sqr(Scal(One)-Z)*LI2(T,S)) / (Scal(Two)+Two*Sqr(Z));
   end Zeta2I;

   function Sigma2(Z: in Scalar) return Scalar is
   begin
      return (Exp(Kappa2*Log(Z+Scal(One)))-Exp(Kappa2*Log(Scal(One)-Z)))  / (Exp(Kappa2*Log(Z+Scal(One)))+Exp(Kappa2*Log(Scal(One)-Z))) ;
   end Sigma2;

   procedure Sigma2(Z1,Z2: in out Scalar)  is
      S1,S2,S,T1,T2,T: Scalar;
   begin
      S1:=(Z1+Scal(One));
      S2:=Z2;
      Ln(S1,S2);
      S1:=S1*Kappa2;
      S2:=S2*Kappa2;
      S:= Exp(S1)*Cos(S2);
      S2:=Exp(S1)*Sin(S2);
      S1:=S;
      ---------------------------------------------
      T1:=(Scal(One)-Z1);
      T2:=-Z2;
      Ln(T1,T2);
      T1:=T1*Kappa2;
      T2:=T2*Kappa2;
      T:= Exp(T1)*Cos(T2);
      T2:=Exp(T1)*Sin(T2);
      T1:=T;
      ---------------------------------------------
      -- ( (S1-T1)+I*(S2-T2) )*( (S1+T1)-I(S2+T2))    / ( Sqr(S1+T1)+Sqr(S2+T2)  );

      Z1:=( (S1-T1)*(S1+T1) + (S2-T2)*(S2+T2) )/( Sqr(S1+T1)+Sqr(S2+T2) );
      Z2:=Two*(S2*T1-S1*T2)/( Sqr(S1+T1)+Sqr(S2+T2) );
   end Sigma2;

   function Sigma2I(Z: in Scalar) return Scalar is
   begin
      return (Exp(Log(Z+Scal(One))/Kappa2)-Exp(Log(Scal(One)-Z)/Kappa2) )  / (Exp(Log(Z+Scal(One))/Kappa2)+Exp(Log(Scal(One)-Z)/Kappa2)) ;
   end Sigma2I;

   function Q2(Z,T,S: in Scalar) return Scalar is
   begin
      return Half*(RI2(T,S)-LI2(T,S))*Z+Half*(RI2(T,S)+LI2(T,S));
   end Q2;

   procedure Q2(Z1,Z2: in out Scalar; T,S: in Scalar) is
   begin

      Z1:=Half*(RI2(T,S)-LI2(T,S))*Z1 + Half*(RI2(T,S)+LI2(T,S));
      Z2:=Half*(RI2(T,S)-LI2(T,S))*Z2;
   end Q2;

   function Q2I(Z,T,S: in Scalar) return Scalar is
   begin
      return (Two*Z-(RI2(T,S)+LI2(T,S)))/(RI2(T,S)-LI2(T,S));
   end Q2I;

   function M2(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma2I(Q2I(Scal(Zero),T,S))-Zeta2(Scal(Zero),T,S)) / (Scal(One)- Sigma2I(Q2I(Scal(Zero),T,S))*Zeta2(Scal(Zero),T,S));
   begin
      return (Z+Ak)/(Scal(One)+Ak*Z);
   end M2;

   procedure M2(Z1,Z2: in out Scalar; T,S: in Scalar) is
      Ak: constant Scalar:=(Sigma2I(Q2I(Scal(Zero),T,S))-Zeta2(Scal(Zero),T,S)) / (Scal(One)- Sigma2I(Q2I(Scal(Zero),T,S))*Zeta2(Scal(Zero),T,S));
      R: Scalar:=Sqr(Scal(One)+Ak*Z1) + Sqr(Ak*Z2);
   begin
      -- (Z1+I*Z2+Ak)/(Scal(One)+Ak*(Z1+I*Z2));
      Z1:=( (Z1+Ak)*(Scal(One)+Ak*Z1)+Ak*Sqr(Z2) )/R;
      Z2:=(Z2-Sqr(Ak)*Z2)/R;
   end M2;

   function M2I(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma2I(Q2I(Scal(Zero),T,S))-Zeta2(Scal(Zero),T,S)) / (Scal(One)- Sigma2I(Q2I(Scal(Zero),T,S))*Zeta2(Scal(Zero),T,S));
   begin
      return (Z-Ak)/(Scal(One)-Ak*Z);
   end M2I;

   function Thet2(Z,T,S: in Scalar) return Scalar is
   begin
      return Q2(Sigma2(M2(Zeta2(Z,T,S),T,S)),T,S) ;
   end Thet2;

   procedure Thet2(Z1,Z2: in out Scalar; T,S: in Scalar) is
   begin
      Zeta2(Z1,Z2,T,S);
      M2(Z1,Z2,T,S);
      Sigma2(Z1,Z2);
      Q2(Z1,Z2,T,S);
   end Thet2;

   function Ph2(Z,T,S: in Scalar) return Scalar is
   begin
      return Zeta2I(M2I(Sigma2I(Q2I(Z,T,S)),T,S),T,S) ;
   end Ph2;

   function Lin2I(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph2(Scal(Zero),T,S);
      C_2: Scalar:=Ph2( Scal(One),T,S);
   begin
      return (C3-C4)*Z/(C_1-C_2)+(C4*C_1-C3*C_2)/(C_1-C_2);

   end Lin2I;

   function Lin2(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph2(Scal(Zero),T,S);
      C_2: Scalar:=Ph2( Scal(One),T,S);
   begin
      return (Z*(C_1-C_2)-C4*C_1+C3*C_2 ) / (C3-C4);
   end Lin2;

   procedure Lin2(Z1,Z2: in out Scalar; T,S: in Scalar) is
      C_1: Scalar:=Ph2(Scal(Zero),T,S);
      C_2: Scalar:=Ph2( Scal(One),T,S);
   begin
      Z1:=(Z1*(C_1-C_2)-C4*C_1+C3*C_2 ) / (C3-C4);
      Z2:= Z2*(C_1-C_2) / (C3-C4);
   end Lin2;

   function Theta2(Z,T,S: in Scalar) return Scalar is
   begin
      return  Thet2(Lin2(Z,T,S),T,S);
   end Theta2;

   procedure Theta2(Z1,Z2: in out Scalar; T,S: in Scalar) is
      ZZ2: Scalar:=Z2;
   begin
      Lin2(Z1,Z2,T,S);
      Thet2(Z1,Z2,T,S);
      if  Inf(ZZ2)>Zero and Sup(Z2)<Zero then
         Message("Doubling.Theta2",Singular);
      elsif Inf(ZZ2)>Zero and Inf(Z2)<Zero then
         Z2:=Scal(Zero,Sup(Z2));
      end if;

      if  Sup(ZZ2)<Zero and Inf(Z2)>Zero then
         Message("Doubling.Teta2",Singular);
      elsif Sup(ZZ2)<Zero and Sup(Z2)>Zero then
         Z2:=Scal(Inf(Z2),Zero);
      end if;
   end Theta2;

   function Phi2(Z,T,S: in Scalar) return Scalar is
   begin
      return Lin2I(Ph2(Z,T,S),T,S);
   end Phi2;

   ---------------------------------------------

   function Zeta3(Z,T,S: in Scalar) return Scalar is
   begin
      return ( Sqrt(Z-LI3(T,S)) - Sqrt(RI3(T,S)-Z) )/(Sqrt(Z-LI3(T,S))+Sqrt(RI3(T,S)-Z));
   end Zeta3;

   procedure Zeta3(Z1,Z2: in out Scalar; T,S: in Scalar) is
      S1,S2,SS,T1,T2,TT: Scalar;
   begin
      S1:=Z1-LI3(T,S);
      S2:=Z2;
      Ln(S1,S2);
      S1:=Half*S1;
      S2:=Half*S2;
      SS:= Exp(S1)*Cos(S2);
      S2:=Exp(S1)*Sin(S2);
      S1:=SS;
      ---------------------------------------------
      T1:=RI3(T,S)-Z1;
      T2:=-Z2;
      Ln(T1,T2);
      T1:=Half*T1;
      T2:=Half*T2;
      TT:= Exp(T1)*Cos(T2);
      T2:=Exp(T1)*Sin(T2);
      T1:=TT;
      ---------------------------------------------
      Z1:=( (S1-T1)*(S1+T1) + (S2-T2)*(S2+T2) )/( Sqr(S1+T1)+Sqr(S2+T2) );
      Z2:=Two*(S2*T1-S1*T2)/( Sqr(S1+T1)+Sqr(S2+T2) );
   end Zeta3;


   function Zeta3I(Z,T,S: in Scalar) return Scalar is
   begin
      return (Sqr(Scal(One)+Z)*RI3(T,S)+Sqr(Scal(One)-Z)*LI3(T,S)) / (Scal(Two)+Two*Sqr(Z));
   end Zeta3I;

   function Sigma3(Z: in Scalar) return Scalar is
   begin
      return (Exp(Kappa3*Log(Z+Scal(One)))-Exp(Kappa3*Log(Scal(One)-Z)))  / (Exp(Kappa3*Log(Z+Scal(One)))+Exp(Kappa3*Log(Scal(One)-Z))) ;
   end Sigma3;

   procedure Sigma3(Z1,Z2: in out Scalar)  is
      S1,S2,S,T1,T2,T: Scalar;
   begin
      S1:=(Z1+Scal(One));
      S2:=Z2;
      Ln(S1,S2);
      S1:=S1*Kappa3;
      S2:=S2*Kappa3;
      S:= Exp(S1)*Cos(S2);
      S2:=Exp(S1)*Sin(S2);
      S1:=S;
      ---------------------------------------------
      T1:=(Scal(One)-Z1);
      T2:=-Z2;
      Ln(T1,T2);
      T1:=T1*Kappa3;
      T2:=T2*Kappa3;
      T:= Exp(T1)*Cos(T2);
      T2:=Exp(T1)*Sin(T2);
      T1:=T;
      ---------------------------------------------
      -- ( (S1-T1)+I*(S2-T2) )*( (S1+T1)-I(S2+T2))    / ( Sqr(S1+T1)+Sqr(S2+T2)  );

      Z1:=( (S1-T1)*(S1+T1) + (S2-T2)*(S2+T2) )/( Sqr(S1+T1)+Sqr(S2+T2) );
      Z2:=Two*(S2*T1-S1*T2)/( Sqr(S1+T1)+Sqr(S2+T2) );
   end Sigma3;


   function Sigma3I(Z: in Scalar) return Scalar is
   begin
      return (Exp(Log(Z+Scal(One))/Kappa3)-Exp(Log(Scal(One)-Z)/Kappa3) )  / (Exp(Log(Z+Scal(One))/Kappa3)+Exp(Log(Scal(One)-Z)/Kappa3)) ;
   end Sigma3I;

   function Q3(Z,T,S: in Scalar) return Scalar is
   begin
      return Half*(RI3(T,S)-LI3(T,S))*Z+Half*(RI3(T,S)+LI3(T,S));
   end Q3;

   procedure Q3(Z1,Z2: in out Scalar; T,S: in Scalar) is
   begin

      Z1:=Half*(RI3(T,S)-LI3(T,S))*Z1+Half*(RI3(T,S)+LI3(T,S));
      Z2:=Half*(RI3(T,S)-LI3(T,S))*Z2;
   end Q3;


   function Q3I(Z,T,S: in Scalar) return Scalar is
   begin
      return (Two*Z-(RI3(T,S)+LI3(T,S)))/(RI3(T,S)-LI3(T,S));
   end Q3I;

   function M3(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma3I(Q3I(Scal(Zero),T,S))-Zeta3(Scal(Zero),T,S)) / (Scal(One)- Sigma3I(Q3I(Scal(Zero),T,S))*Zeta3(Scal(Zero),T,S));
   begin
      return (Z+Ak)/(Scal(One)+Ak*Z);
   end M3;

   procedure M3(Z1,Z2: in out Scalar; T,S: in Scalar) is
      Ak: constant Scalar:=(Sigma3I(Q3I(Scal(Zero),T,S))-Zeta3(Scal(Zero),T,S)) / (Scal(One)- Sigma3I(Q3I(Scal(Zero),T,S))*Zeta3(Scal(Zero),T,S));
      R: Scalar:=Sqr(Scal(One)+Ak*Z1) + Sqr(Ak*Z2);
   begin
      -- (Z1+I*Z2+Ak)/(Scal(One)+Ak*(Z1+I*Z2));
      Z1:=( (Z1+Ak)*(Scal(One)+Ak*Z1)+Ak*Sqr(Z2) )/R;
      Z2:=(Z2-Sqr(Ak)*Z2)/R;
   end M3;

   function M3I(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma3I(Q3I(Scal(Zero),T,S))-Zeta3(Scal(Zero),T,S)) / (Scal(One)- Sigma3I(Q3I(Scal(Zero),T,S))*Zeta3(Scal(Zero),T,S));
   begin
      return (Z-Ak)/(Scal(One)-Ak*Z);
   end M3I;

   function Thet3(Z,T,S: in Scalar) return Scalar is
   begin
      return Q3(Sigma3(M3(Zeta3(Z,T,S),T,S)),T,S) ;
   end Thet3;

   procedure Thet3(Z1,Z2: in out Scalar; T,S: in Scalar) is
   begin
      Zeta3(Z1,Z2,T,S);
      M3(Z1,Z2,T,S);
      Sigma3(Z1,Z2);
      Q3(Z1,Z2,T,S);
   end Thet3;


   function Ph3(Z,T,S: in Scalar) return Scalar is
   begin
      return Zeta3I(M3I(Sigma3I(Q3I(Z,T,S)),T,S),T,S);
   end Ph3;

   function Lin3(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph3(Scal(Zero),T,S);
      C_2: Scalar:=Ph3( Scal(One),T,S);
   begin
      return (Z*(C_1-C_2)-(C4*C_1-C3*C_2) ) / (C3-C4);
   end Lin3;

   procedure Lin3(Z1,Z2: in out Scalar; T,S: in Scalar) is
      C_1: Scalar:=Ph3(Scal(Zero),T,S);
      C_2: Scalar:=Ph3( Scal(One),T,S);
   begin
      Z1:=(Z1*(C_1-C_2)-C4*C_1+C3*C_2 ) / (C3-C4);
      Z2:=Z2*(C_1-C_2) / (C3-C4);
   end Lin3;


   function Lin3I(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph3(Scal(Zero),T,S);
      C_2: Scalar:=Ph3( Scal(One),T,S);
   begin
      return (C3-C4)*Z/(C_1-C_2)+(C4*C_1-C3*C_2)/(C_1-C_2);

   end Lin3I;

   function Theta3(Z,T,S: in Scalar) return Scalar is
   begin
      return  Thet3(Lin3(Z,T,S),T,S);
   end Theta3;

   procedure Theta3(Z1,Z2: in out Scalar; T,S: in Scalar) is
      ZZ2: Scalar:=Z2;
   begin
      Lin3(Z1,Z2,T,S);
      Thet3(Z1,Z2,T,S);
      if  Inf(ZZ2)>Zero and Sup(Z2)<Zero then
         Message("Doubling.Theta3",Singular);
      elsif Inf(ZZ2)>Zero and Inf(Z2)<Zero then
         Z2:=Scal(Zero,Sup(Z2));
      end if;

      if  Sup(ZZ2)<Zero and Inf(Z2)>Zero then
         Message("Doubling.Theta3",Singular);
      elsif Sup(ZZ2)<Zero and Sup(Z2)>Zero then
         Z2:=Scal(Inf(Z2),Zero);
      end if;
   end Theta3;

   function Phi3(Z,T,S: in Scalar) return Scalar is
   begin
      return Lin3I(Ph3(Z,T,S),T,S);
   end Phi3;

   ---------------------------------------------

   function DerZeta1(Z: in Scalar) return Scalar is
   begin
      return Scal(Two)/(Sqr(Sqrt(Z+Scal(One))+Sqrt(Scal(One)-Z))*Sqrt(Z+Scal(One))*Sqrt(Scal(One)-Z));
   end DerZeta1;

   function DerZeta1I(Z: in Scalar) return Scalar is
   begin
      return
        -Two*(Sqr(Z)-Scal(One))/Sqr(Sqr(Z)+Scal(One));
   end DerZeta1I;

   function DerSigma1(Z: in Scalar) return Scalar is
   begin
      return  Two*Kappa1*( Exp((Kappa1-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa1*Log(Scal(One)-Z)) + Exp((Kappa1-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa1*Log(Scal(One)+Z)))/Sqr( Exp(Kappa1*Log(Z+Scal(One)))+ Exp(Kappa1*Log(Scal(One)-Z)));
   end DerSigma1;

   function DerSigma1I(Z: in Scalar) return Scalar is
      Kappa: constant Scalar:=Scal(One)/Kappa1;
   begin
      return  Two*Kappa*( Exp((Kappa-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa*Log(Scal(One)-Z)) + Exp((Kappa-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa*Log(Scal(One)+Z)))/Sqr( Exp(Kappa*Log(Z+Scal(One)))+ Exp(Kappa*Log(Scal(One)-Z)));
   end DerSigma1I;

   function DerM1(Z: in Scalar) return Scalar is
      AA: constant Scalar:=Sigma1I(Q1I(-Scal(Half)));
   begin
      return -(Sqr(AA)-Scal(One))/Sqr(Scal(One)+AA*Z);
   end DerM1;

   function DerM1I(Z: in Scalar) return Scalar is
      AA: constant Scalar:=Sigma1I(Q1I(-Scal(Half)));
   begin
      return -(Sqr(AA)-Scal(One))/Sqr(Scal(One)-AA*Z);
   end DerM1I;

   function DerQ1(Z: in Scalar) return Scalar is
   begin
      return Half*(R1-L1);
   end DerQ1;

   function DerQ1I(Z: in Scalar) return Scalar is
   begin
      return Scal(Two)/(R1-L1);
   end DerQ1I;

   function DerPhi1(Z: in Scalar) return Scalar is
      Z1: Scalar:=Q1I(Z);
      Z2: Scalar:=Sigma1I(Z1);
      Z3: Scalar:=M1I(Z2);
   begin
      return
        DerZeta1I(Z3)*
        DerM1I(Z2)*
        DerSigma1I(Z1)*
        DerQ1I(Z);
   end DerPhi1;

   ---------------------------------------------

   function DerZeta2(Z,T,S: in Scalar) return Scalar is
   begin
      return (RI2(T,S)-LI2(T,S))/(Sqr(Sqrt(Z-LI2(T,S))+Sqrt(RI2(T,S)-Z))*Sqrt(Z-LI2(T,S))*Sqrt(RI2(T,S)-Z));
   end DerZeta2;

   function DerZeta2I(Z,T,S: in Scalar) return Scalar is
   begin
      return (LI2(T,S)-RI2(T,S))*(Sqr(Z)-Scal(One))/Sqr(Sqr(Z)+Scal(One));
   end DerZeta2I;

   function DerSigma2(Z: in Scalar) return Scalar is
   begin
      return  Two*Kappa2*( Exp((Kappa2-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa2*Log(Scal(One)-Z)) + Exp((Kappa2-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa2*Log(Scal(One)+Z)))/Sqr( Exp(Kappa2*Log(Z+Scal(One)))+ Exp(Kappa2*Log(Scal(One)-Z)));
   end DerSigma2;

   function DerSigma2I(Z: in Scalar) return Scalar is
      Kappa: constant Scalar:=Scal(One)/Kappa2;
   begin
      return  Two*Kappa*( Exp((Kappa-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa*Log(Scal(One)-Z)) + Exp((Kappa-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa*Log(Scal(One)+Z)))/Sqr( Exp(Kappa*Log(Z+Scal(One)))+ Exp(Kappa*Log(Scal(One)-Z)));
   end DerSigma2I;

   function DerM2(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma2I(Q2I(Scal(Zero),T,S))-Zeta2(Scal(Zero),T,S)) / (Scal(One)- Sigma2I(Q2I(Scal(Zero),T,S))*Zeta2(Scal(Zero),T,S));
   begin
      return -(Sqr(Ak)-Scal(One))/Sqr(Scal(One)+Ak*Z);
   end DerM2;

   function DerM2I(Z,T,S: in Scalar) return Scalar is
      Ak: constant Scalar:=(Sigma2I(Q2I(Scal(Zero),T,S))-Zeta2(Scal(Zero),T,S)) / (Scal(One)- Sigma2I(Q2I(Scal(Zero),T,S))*Zeta2(Scal(Zero),T,S));
   begin
      return -(Sqr(Ak)-Scal(One))/Sqr(Scal(One)-Ak*Z);
   end DerM2I;

   function DerQ2(Z,T,S: in Scalar) return Scalar is
   begin
      return Half*(RI2(T,S)-LI2(T,S));
   end DerQ2;

   function DerQ2I(Z,T,S: in Scalar) return Scalar is
   begin
      return Scal(Two)/(RI2(T,S)-LI2(T,S));
   end DerQ2I;

   function DerLin2(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph2(Scal(Zero),T,S);
      C_2: Scalar:=Ph2( Scal(One),T,S);
   begin
      return (C_1-C_2)/ (C3-C4);
   end DerLin2;

   function DerLin2I(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph2(Scal(Zero),T,S);
      C_2: Scalar:=Ph2( Scal(One),T,S);
   begin
      return (C3-C4)/(C_1-C_2);
   end DerLin2I;

   function DerTheta2(Z,T,S: in Scalar) return Scalar is
      Z1: Scalar:=Lin2(Z,T,S);
      Z2: Scalar:=Zeta2(Z1,T,S);
      Z3: Scalar:=M2(Z2,T,S);
      Z4: Scalar:=Sigma2(Z3);
   begin
      return
        DerQ2(Z4,T,S)*
        DerSigma2(Z3)*
        DerM2(Z2,T,S)*
        DerZeta2(Z1,T,S)*
        DerLin2(Z,T,S);
   end DerTheta2;

   ---------------------------------------------

   function DerZeta3(Z,T,S: in Scalar) return Scalar is
   begin
      return (RI3(T,S)-LI3(T,S))/(Sqr(Sqrt(Z-LI3(T,S))+Sqrt(RI3(T,S)-Z))*Sqrt(Z-LI3(T,S))*Sqrt(RI3(T,S)-Z));
   end DerZeta3;

   function DerZeta3I(Z,T,S: in Scalar) return Scalar is
   begin
      return (LI3(T,S)-RI3(T,S))*(Sqr(Z)-Scal(One))/Sqr(Sqr(Z)+Scal(One));
   end DerZeta3I;

   function DerSigma3(Z: in Scalar) return Scalar is
   begin
      return  Two*Kappa3*( Exp((Kappa3-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa3*Log(Scal(One)-Z)) + Exp((Kappa3-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa3*Log(Scal(One)+Z)))/Sqr( Exp(Kappa3*Log(Z+Scal(One)))+ Exp(Kappa3*Log(Scal(One)-Z)));
   end DerSigma3;

   function DerSigma3I(Z: in Scalar) return Scalar is
      Kappa: constant Scalar:=Scal(One)/Kappa3;
   begin
      return  Two*Kappa*( Exp((Kappa-Scal(One))*Log(Z+Scal(One)))*Exp(Kappa*Log(Scal(One)-Z)) + Exp((Kappa-Scal(One))*Log(Scal(One)-Z))*Exp(Kappa*Log(Scal(One)+Z)))/Sqr( Exp(Kappa*Log(Z+Scal(One)))+ Exp(Kappa*Log(Scal(One)-Z)));
   end DerSigma3I;

   function DerM3(Z,T,S: in Scalar) return Scalar is
      AA: constant Scalar:=( Sigma3I(Q3I(Scal(Zero),T,S))-Zeta3(Scal(Zero),T,S) ) /(Scal(One)-Zeta3(Scal(Zero),T,S)*Sigma3I(Q3I(Scal(Zero),T,S))) ;
   begin
      return -(Sqr(AA)-Scal(One))/Sqr(Scal(One)+AA*Z);
   end DerM3;

   function DerM3I(Z,T,S: in Scalar) return Scalar is
      AA: constant Scalar:=( Sigma3I(Q3I(Scal(Zero),T,S))-Zeta3(Scal(Zero),T,S) ) /(Scal(One)-Zeta3(Scal(Zero),T,S)*Sigma3I(Q3I(Scal(Zero),T,S))) ;
   begin
      return -(Sqr(AA)-Scal(One))/Sqr(Scal(One)-AA*Z);
   end DerM3I;

   function DerQ3(Z,T,S: in Scalar) return Scalar is
   begin
      return Half*(RI3(T,S)-LI3(T,S));
   end DerQ3;

   function DerQ3I(Z,T,S: in Scalar) return Scalar is
   begin
      return Scal(Two)/(RI3(T,S)-LI3(T,S));
   end DerQ3I;

   function DerLin3(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph3(Scal(Zero),T,S);
      C_2: Scalar:=Ph3(Scal(One),T,S);
   begin
      return (C_1-C_2)/ (C3-C4);
   end DerLin3;

   function DerLin3I(Z,T,S: in Scalar) return Scalar is
      C_1: Scalar:=Ph3(Scal(Zero),T,S);
      C_2: Scalar:=Ph3(Scal(One),T,S);
   begin
      return (C3-C4)/(C_1-C_2);
   end DerLin3I;

   function DerTheta3(Z,T,S: in Scalar) return Scalar is
      Z1: Scalar:=Lin3(Z,T,S);
      Z2: Scalar:=Zeta3(Z1,T,S);
      Z3: Scalar:=M3(Z2,T,S);
      Z4: Scalar:=Sigma3(Z3);
   begin
      return
        DerQ3(Z4,T,S)*
        DerSigma3(Z3)*
        DerM3(Z2,T,S)*
        DerZeta3(Z1,T,S)*
        DerLin3(Z,T,S);
   end DerTheta3;

   ---------------------------------------------

   function T2(T,S: in Scalar) return Scalar is
   begin
      return T/(DerTheta2(C3,T,S)*DerPhi1(-Scal(Half)));
   end T2;

   function T3(T,S: in Scalar) return Scalar is
   begin
      return T/(DerTheta3(C3,T,S)*DerPhi1(-Scal(Half)));
   end T3;

   function S2(T,S: in Scalar) return Scalar is
   begin
      return S/(DerTheta2(C4,T,S)*DerPhi1(Scal(Zero)));
   end S2;

   function S3(T,S: in Scalar) return Scalar is
   begin
      return S/(DerTheta3(C4,T,S)*DerPhi1(Scal(Zero)));
   end S3;

   function Z2(T,S: in Scalar) return Scalar is
      C_1: Scalar:=Phi1(Scal(Zero));
      C_2: Scalar:=Phi1(Scal(-Half));
      S_2: constant Scalar:=S2(T,S);
      T_2: constant Scalar:=T2(T,S);
   begin
      return (Two*S_2-Two*Sqrt(S_2*T_2))*T_2/((C_1-C_2)*S_2);
   end Z2;

   function Z3(T,S: in Scalar) return Scalar is
      C_1: Scalar:=Phi1(Scal(Zero));
      C_2: Scalar:=Phi1(Scal(-Half));
      S_3: constant Scalar:=S3(T,S);
      T_3: constant Scalar:=T3(T,S);
   begin
      return (Two*S_3-Two*Sqrt(S_3*T_3))*T_3/((C_1-C_2)*S_3);
   end Z3;

   function X2(T,S: in Scalar) return Scalar is
      C_1: Scalar:=Phi1(Scal(Zero));
      C_2: Scalar:=Phi1(Scal(-Half));
      S_2: constant Scalar:=S2(T,S);
      T_2: constant Scalar:=T2(T,S);
   begin
      return (Two*T_2-Two*Sqrt(S_2*T_2))*S_2/((C_2-C_1)*T_2);
   end X2;

   function X3(T,S: in Scalar) return Scalar is
      C_1: Scalar:=Phi1(Scal(Zero));
      C_2: Scalar:=Phi1(Scal(-Half));
      S_3: constant Scalar:=S3(T,S);
      T_3: constant Scalar:=T3(T,S);
   begin
      return (Two*T_3-Two*Sqrt(S_3*T_3))*S_3/((C_2-C_1)*T_3);
   end X3;

   ---------------------------------------------

   function MS2(T,S: in Scalar) return Scalar is
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return Two*S2(T,S)/(Scal(One)-C2);
   end MS2;

   function MZ2(T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return (-Two)*S2(T,S)/(Scal(One)+C1);
   end MZ2;

   function MS3(T,S: in Scalar) return Scalar is
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return Two*S3(T,S)/(Scal(One)-C2);
   end MS3;

   function MZ3(T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return (-Two)*S3(T,S)/(Scal(One)+C1);
   end MZ3;

   function MT2(T,S: in Scalar) return Scalar is
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return Two*T2(T,S)/(Scal(One)-C2);
   end MT2;

   function MD2(T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return (-Two)*T2(T,S)/(Scal(One)+C1);
   end MD2;

   function MT3(T,S: in Scalar) return Scalar is
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return Two*T3(T,S)/(Scal(One)-C2);
   end MT3;

   function MD3(T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return (-Two)*T3(T,S)/(Scal(One)+C1);
   end MD3;

   function XD2(T,S,Q: in Scalar) return Scalar is
      M_2: Scalar:=MD2(T,S);
      C1: Scalar:=Phi1(Scal(-Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (T2(T,S)-S2(T,S)+Q*C2-M_2*C1)/(Q-M_2);
   end XD2;

   function XZ2(T,S,Q: in Scalar) return Scalar is
      M_2: Scalar:=MZ2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (T2(T,S)-S2(T,S)+Q*C2-M_2*C1)/(Q-M_2);
   end XZ2;

   function XT2(T,S,Q: in Scalar) return Scalar is
      M_2: Scalar:=MT2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (S2(T,S)-T2(T,S)+Q*C1-M_2*C2)/(Q-M_2);
   end XT2;

   function XS2(T,S,Q: in Scalar) return Scalar is
      M_2: Scalar:=MS2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (S2(T,S)-T2(T,S)+Q*C1-M_2*C2)/(Q-M_2);
   end XS2;

   function XD3(T,S,Q: in Scalar) return Scalar is
      M_3: Scalar:=MD3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (T3(T,S)-S3(T,S)+Q*C2-M_3*C1)/(Q-M_3);
   end XD3;

   function XZ3(T,S,Q: in Scalar) return Scalar is
      M_3: Scalar:=MZ3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (T3(T,S)-S3(T,S)+Q*C2-M_3*C1)/(Q-M_3);
   end XZ3;

   function XT3(T,S,Q: in Scalar) return Scalar is
      M_3: Scalar:=MT3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (S3(T,S)-T3(T,S)+Q*C1-M_3*C2)/(Q-M_3);
   end XT3;

   function XS3(T,S,Q: in Scalar) return Scalar is
      M_3: Scalar:=MS3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (S3(T,S)-T3(T,S)+Q*C1-M_3*C2)/(Q-M_3);
   end XS3;

   function NT2(T,S: in Scalar) return Scalar is
      T_2: Scalar:=T2(T,S);
      S_2: Scalar:=S2(T,S);
      M_2: Scalar:=MT2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return  (Sqr(T_2)-Two*T_2*C1*M_2+Sqr(S_2)-Two*S_2*T_2+Two*T_2*M_2*C2+Two*C3*M_2-Two*C4*M_2)/(-Sqr(C1)*M_2+Two*C1*M_2*C2+Two*C3-M_2*Sqr(C2)-Two*C4+Two*S_2*C2-Two*S_2*C1);
   end NT2;

   function NS2(T,S: in Scalar) return Scalar is
      T_2: Scalar:=T2(T,S);
      S_2: Scalar:=S2(T,S);
      M_2: Scalar:=MS2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (Sqr(T_2)-Two*T_2*C1*M_2+Sqr(S_2)-Two*S_2*T_2+Two*T_2*M_2*C2+Two*C3*M_2-Two*C4*M_2)/(-Sqr(C1)*M_2+Two*C1*M_2*C2+Two*C3-M_2*Sqr(C2)-Two*C4+Two*S_2*C2-Two*S_2*C1);
   end NS2;

   function ND2(T,S: in Scalar) return Scalar is
      T_2: Scalar:=T2(T,S);
      S_2: Scalar:=S2(T,S);
      M_2: Scalar:=MD2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return -(Two*S_2*M_2*C1+Sqr(S_2)-Two*S_2*C2*M_2-Two*S_2*T_2+M_2+Sqr(T_2))/(M_2*Sqr(C1)-Two*C2*M_2*C1+Sqr(C2)*M_2-Scal(One)+Two*T_2*C2-Two*T_2*C1);
   end ND2;

   function NZ2(T,S: in Scalar) return Scalar is
      T_2: Scalar:=T2(T,S);
      S_2: Scalar:=S2(T,S);
      M_2: Scalar:=MZ2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return -(Two*S_2*M_2*C1+Sqr(S_2)-Two*S_2*C2*M_2-Two*S_2*T_2+M_2+Sqr(T_2))/(M_2*Sqr(C1)-Two*C2*M_2*C1+Sqr(C2)*M_2-Scal(One)+Two*T_2*C2-Two*T_2*C1);
   end NZ2;

   function NT3(T,S: in Scalar) return Scalar is
      T_3: Scalar:=T3(T,S);
      S_3: Scalar:=S3(T,S);
      M_3: Scalar:=MT3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return  (Sqr(T_3)-Two*T_3*C1*M_3+Sqr(S_3)-Two*S_3*T_3+Two*T_3*M_3*C2+Two*C3*M_3-Two*C4*M_3)/(-Sqr(C1)*M_3+Two*C1*M_3*C2+Two*C3-M_3*Sqr(C2)-Two*C4+Two*S_3*C2-Two*S_3*C1);
   end NT3;

   function NS3(T,S: in Scalar) return Scalar is
      T_3: Scalar:=T3(T,S);
      S_3: Scalar:=S3(T,S);
      M_3: Scalar:=MS3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return (Sqr(T_3)-Two*T_3*C1*M_3+Sqr(S_3)-Two*S_3*T_3+Two*T_3*M_3*C2+Two*C3*M_3-Two*C4*M_3)/(-Sqr(C1)*M_3+Two*C1*M_3*C2+Two*C3-M_3*Sqr(C2)-Two*C4+Two*S_3*C2-Two*S_3*C1);
   end NS3;

   function ND3(T,S: in Scalar) return Scalar is
      T_3: Scalar:=T3(T,S);
      S_3: Scalar:=S3(T,S);
      M_3: Scalar:=MD3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return -(Two*S_3*M_3*C1+Sqr(S_3)-Two*S_3*C2*M_3-Two*S_3*T_3+M_3+Sqr(T_3))/(M_3*Sqr(C1)-Two*C2*M_3*C1+Sqr(C2)*M_3-Scal(One)+Two*T_3*C2-Two*T_3*C1);
   end ND3;

   function NZ3(T,S: in Scalar) return Scalar is
      T_3: Scalar:=T3(T,S);
      S_3: Scalar:=S3(T,S);
      M_3: Scalar:=MZ3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return -(Two*S_3*M_3*C1+Sqr(S_3)-Two*S_3*C2*M_3-Two*S_3*T_3+M_3+Sqr(T_3))/(M_3*Sqr(C1)-Two*C2*M_3*C1+Sqr(C2)*M_3-Scal(One)+Two*T_3*C2-Two*T_3*C1);
   end NZ3;

   -----------------------------------------------

   --CORRECT
   function U12(Z,T,S: in Scalar) return Scalar is
      T_2: Scalar:=T2(T,S);
      Z_2: Scalar:=Z2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return C3-Scal(4.0)*Sqr(T_2)*T_2*(Scal(Half)/T_2 -Scal(One)/(Two*T_2+Z_2*(C1-Z)))/Z_2;
   end U12;
   --CORRECT
   function U22(Z,T,S: in Scalar) return Scalar is
      N_S2: constant Scalar:=NS2(T,S);
      N_T2: constant Scalar:=NT2(T,S);
      M_S2: constant Scalar:=MS2(T,S);
      M_T2: constant Scalar:=MT2(T,S);
      X_S2: constant Scalar:=XS2(T,S,N_S2);
      X_T2: constant Scalar:=XT2(T,S,N_T2);

      T_2: constant Scalar:=T2(T,S);
      S_2: constant Scalar:=S2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Inf(S_2)>Sup(T_2) then
         if Inf(X_S2)>Sup(Z) then
            return C4+(T_2*(Z-X_S2)+Half*N_S2*(Sqr(Z-C1)-Sqr(X_S2-C1))+S_2*(X_S2-C2)+Half*M_S2*Sqr(X_S2-C2));
         elsif Sup(X_S2)<Inf(Z) then
            return C4+(S_2*(Z-C2)+Half*M_S2*Sqr(Z-C2));
         else
            Message("Doubling.U22",Singular);
            return C4;
         end if;
      elsif Sup(S_2)<Inf(T_2) then
         if Inf(X_T2)>Sup(Z) then
            return C4+(T_2*(Z-X_T2)+Half*N_T2*(Sqr(Z-C1)-Sqr(X_T2-C1))+S_2*(X_T2-C2)+Half*M_T2*Sqr(X_T2-C2));
         elsif Sup(X_T2)<Inf(Z) then
            return C4+(S_2*(Z-C2)+Half*M_T2*Sqr(Z-C2));
         else
            Message("Doubling.U22",Singular);
            return C4;
         end if;
      else
         Message("Doubling.U22",Singular);
         return C4;
      end if;
   end U22;

   --CORRECT
   function U32(Z,T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin2I(RI2(T,S),T,S);
      S_2: constant Scalar:=S2(T,S);
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Sup(RI3(T,S))<Sup(RI2(T,S)) then
         UM:=Phi2(RI3(T,S),T,S);
      end if;
      return C4+S_2*(Z-C2)+(UM-C4-S_2*(Scal(One)-C2) )*Sqr(Z-C2)/Sqr(Scal(One)-C2);
   end U32;

   -----------------------------------------------

   --CORRECT
   function U13(Z,T,S: in Scalar) return Scalar is
      T_2: Scalar:=T3(T,S);
      Z_2: Scalar:=Z3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      return C3-Scal(4.0)*Sqr(T_2)*T_2*(Scal(Half)/T_2 -Scal(One)/(Two*T_2+Z_2*(C1-Z)))/Z_2;
   end U13;
   --CORRECT
   function U23(Z,T,S: in Scalar) return Scalar is
      N_S2: constant Scalar:=NS3(T,S);
      N_T2: constant Scalar:=NT3(T,S);
      M_S2: constant Scalar:=MS3(T,S);
      M_T2: constant Scalar:=MT3(T,S);
      X_S2: constant Scalar:=XS3(T,S,N_S2);
      X_T2: constant Scalar:=XT3(T,S,N_T2);

      T_2: constant Scalar:=T3(T,S);
      S_2: constant Scalar:=S3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Inf(S_2)>Sup(T_2) then
         if Inf(X_S2)>Sup(Z) then
            return C4+(T_2*(Z-X_S2)+Half*N_S2*(Sqr(Z-C1)-Sqr(X_S2-C1))+S_2*(X_S2-C2)+Half*M_S2*Sqr(X_S2-C2));
         elsif Sup(X_S2)<Inf(Z) then
            return C4+(S_2*(Z-C2)+Half*M_S2*Sqr(Z-C2));
         else
            Message("Doubling.U22",Singular);
            return C4;
         end if;
      elsif Sup(S_2)<Inf(T_2) then
         if Inf(X_T2)>Sup(Z) then
            return C4+(T_2*(Z-X_T2)+Half*N_T2*(Sqr(Z-C1)-Sqr(X_T2-C1))+S_2*(X_T2-C2)+Half*M_T2*Sqr(X_T2-C2));
         elsif Sup(X_T2)<Inf(Z) then
            return C4+(S_2*(Z-C2)+Half*M_T2*Sqr(Z-C2));
         else
            Message("Doubling.U22",Singular);
            return C4;
         end if;
      else
         Message("Doubling.U22",Singular);
         return C4;
      end if;
   end U23;

   --CORRECT
   function U33(Z,T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin3I(RI3(T,S),T,S);
      S_2: constant Scalar:=S3(T,S);
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Sup(RI2(T,S))<Sup(RI3(T,S)) then
         UM:=Phi3(RI2(T,S),T,S);
      end if;
      return C4+S_2*(Z-C2)+(UM-C4-S_2*(Scal(One)-C2) )*Sqr(Z-C2)/Sqr(Scal(One)-C2);
   end U33;

   -----------------------------------------------

   function L12(Z,T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin2I(LI2(T,S),T,S);
      T_2: constant Scalar:=T2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      if Inf(LI3(T,S))>Sup(LI2(T,S)) then
         UM:=Phi2(LI3(T,S),T,S);
      end if;
      return C3-T_2*(C1-Z)+(UM-C3+T_2*(Scal(One)+C1) )*Sqr(Z-C1)/Sqr(Scal(One)+C1);
   end L12;

   function L22(Z,T,S: in Scalar) return Scalar is
      N_S2: constant Scalar:=NZ2(T,S);
      N_T2: constant Scalar:=ND2(T,S);
      M_T2: constant Scalar:=MD2(T,S);
      M_S2: constant Scalar:=MZ2(T,S);
      X_S2: constant Scalar:=XZ2(T,S,N_S2);
      X_T2: constant Scalar:=XD2(T,S,N_T2);

      T_2: constant Scalar:=T2(T,S);
      S_2: constant Scalar:=S2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Inf(S_2)>Sup(T_2) then
         if Inf(X_S2)>Sup(Z) then
            return C3+T_2*(Z-C1)+Half*M_S2*Sqr(Z-C1);
         elsif Sup(X_S2)<Inf(Z) then
            return C3+T_2*(X_S2-C1)+Half*M_S2*Sqr(X_S2-C1)+S_2*(Z-X_S2)+Half*N_S2*(Sqr(Z-C2)-Sqr(X_S2-C2));
         else
            Message("Doubling.L22",Singular);
            return C3;
         end if;
      elsif Sup(S_2)<Inf(T_2) then
         if Inf(X_T2)>Sup(Z) then
            return C3+(T_2*(Z-C1)+Half*M_T2*Sqr(Z-C1));
         elsif Sup(X_T2)<Inf(Z) then
            return C3+T_2*(X_T2-C1)+Half*M_T2*Sqr(X_T2-C1)+S_2*(Z-X_T2)+Half*N_T2*(Sqr(Z-C2)-Sqr(X_T2-C2));
         else
            Message("Doubling.L22",Singular);
            return C3;
         end if;
      else
         Message("Doubling.L22",Singular);
         return C3;
      end if;
   end L22;

   function L32(Z,T,S: in Scalar) return Scalar is
      S_2: Scalar:=S2(T,S);
      X_2: Scalar:=X2(T,S);
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return C4+Scal(4.0)*Sqr(S_2)*S_2*(Scal(-Half)/S_2 +Scal(One)/(Two*S_2+X_2*(C2-Z)))/X_2;
   end L32;


   function L13(Z,T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin3I(LI3(T,S),T,S);
      T_2: constant Scalar:=T3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      if Inf(LI2(T,S))>Sup(LI3(T,S)) then
         UM:=Phi3(LI2(T,S),T,S);
      end if;
      return C3-T_2*(C1-Z)+(UM-C3+T_2*(Scal(One)+C1) )*Sqr(Z-C1)/Sqr(Scal(One)+C1);
   end L13;

   function L23(Z,T,S: in Scalar) return Scalar is
      N_S2: constant Scalar:=NZ3(T,S);
      N_T2: constant Scalar:=ND3(T,S);
      M_T2: constant Scalar:=MD3(T,S);
      M_S2: constant Scalar:=MZ3(T,S);
      X_S2: constant Scalar:=XZ3(T,S,N_S2);
      X_T2: constant Scalar:=XD3(T,S,N_T2);

      T_2: constant Scalar:=T3(T,S);
      S_2: constant Scalar:=S3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Inf(S_2)>Sup(T_2) then
         if Inf(X_S2)>Sup(Z) then
            return C3+T_2*(Z-C1)+Half*M_S2*Sqr(Z-C1);
         elsif Sup(X_S2)<Inf(Z) then
            return C3+T_2*(X_S2-C1)+Half*M_S2*Sqr(X_S2-C1)+S_2*(Z-X_S2)+Half*N_S2*(Sqr(Z-C2)-Sqr(X_S2-C2));
         else
            Message("Doubling.L22",Singular);
            return C3;
         end if;
      elsif Sup(S_2)<Inf(T_2) then
         if Inf(X_T2)>Sup(Z) then
            return C3+(T_2*(Z-C1)+Half*M_T2*Sqr(Z-C1));
         elsif Sup(X_T2)<Inf(Z) then
            return C3+T_2*(X_T2-C1)+Half*M_T2*Sqr(X_T2-C1)+S_2*(Z-X_T2)+Half*N_T2*(Sqr(Z-C2)-Sqr(X_T2-C2));
         else
            Message("Doubling.L22",Singular);
            return C3;
         end if;
      else
         Message("Doubling.L22",Singular);
         return C3;
      end if;
   end L23;

   function L33(Z,T,S: in Scalar) return Scalar is
      S_2: Scalar:=S3(T,S);
      X_2: Scalar:=X3(T,S);
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      return C4+Scal(4.0)*Sqr(S_2)*S_2*(Scal(-Half)/S_2 +Scal(One)/(Two*S_2+X_2*(C2-Z)))/X_2;
   end L33;

   -----------------------------------------------
    function Upp_2(Z,T,S: in Scalar) return Scalar is
       CC: Scalar:=Scal(Zero);
       C1: Scalar:=Phi1(-Scal(Half));
       C2: Scalar:=Phi1(Scal(Zero));
    begin
       if Inf(C1)>Sup(Z) then
          CC:=U12(Z,T,S);
       elsif Inf(Z)>Sup(C2) then
          CC:=U32(Z,T,S);
       elsif Sup(C1)<Inf(Z) and Sup(Z)<Inf(C2) then
          CC:=U22(Z,T,S);
       else
          Message("Doubling.Upp22",Singular);
       end if;
       return CC;
    end Upp_2;

    function Upp_3(Z,T,S: in Scalar) return Scalar is
       CC: Scalar:=Scal(Zero);
       C1: Scalar:=Phi1(-Scal(Half));
       C2: Scalar:=Phi1(Scal(Zero));
    begin
       if Inf(C1)>Sup(Z) then
          CC:=U13(Z,T,S);
       elsif Inf(Z)>Sup(C2) then
          CC:=U33(Z,T,S);
       elsif Sup(C1)<Inf(Z) and Sup(Z)<Inf(C2) then
          CC:=U23(Z,T,S);
       else
          Message("Doubling.Upp22",Singular);
       end if;
       return CC;
    end Upp_3;

    function Lou_2(Z,T,S: in Scalar) return Scalar is
       CC: Scalar:=Scal(Zero);
       C1: Scalar:=Phi1(-Scal(Half));
       C2: Scalar:=Phi1(Scal(Zero));
    begin
       if Inf(C1)>Sup(Z) then
          CC:=L12(Z,T,S);
       elsif Inf(Z)>Sup(C2) then
          CC:=L32(Z,T,S);
       elsif Sup(C1)<Inf(Z) and Sup(Z)<Inf(C2) then
          CC:=L22(Z,T,S);
       else
          Message("Doubling.Upp22",Singular);
       end if;
       return CC;
    end Lou_2;

    function Lou_3(Z,T,S: in Scalar) return Scalar is
       CC: Scalar:=Scal(Zero);
       C1: Scalar:=Phi1(-Scal(Half));
       C2: Scalar:=Phi1(Scal(Zero));
    begin
       if Inf(C1)>Sup(Z) then
          CC:=L13(Z,T,S);
       elsif Inf(Z)>Sup(C2) then
          CC:=L33(Z,T,S);
       elsif Sup(C1)<Inf(Z) and Sup(Z)<Inf(C2) then
          CC:=L23(Z,T,S);
       else
          Message("Doubling.Upp22",Singular);
       end if;
       return CC;
    end Lou_3;

    function Upp2(Z,T,S: in Scalar) return Scalar is
    begin
       return Theta2(Upp_2(Phi1(Z),T,S),T,S);
    end Upp2;

    function Upp3(Z,T,S: in Scalar) return Scalar is
    begin
       return Theta3(Upp_3(Phi1(Z),T,S),T,S);
    end Upp3;

    function Lou2(Z,T,S: in Scalar) return Scalar is
    begin
       return Theta2(Lou_2(Phi1(Z),T,S),T,S);
    end Lou2;

    function Lou3(Z,T,S: in Scalar) return Scalar is
    begin
       return Theta3(Lou_3(Phi1(Z),T,S),T,S);
    end Lou3;

    function Upp(Z,T,S: in Scalar) return Scalar is
    begin
       if  Sup(Upp2(Z,T,S))<Inf(Upp3(Z,T,S))  then
          return Upp2(Z,T,S);
       else
          return Upp3(Z,T,S);
       end if;
    end Upp;

    function Lou(Z,T,S: in Scalar) return Scalar is
    begin
       if  Inf(Lou2(Z,T,S))>Sup(Lou3(Z,T,S))  then
          return Lou2(Z,T,S);
       else
          return Lou3(Z,T,S);
       end if;
    end Lou;


   -----------------------------------------------
   function Func2(T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin2I(LI2(T,S),T,S);
      T_2: constant Scalar:=T2(T,S);
      Z_2: constant Scalar:=Z2(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      if Inf(LI3(T,S))>Sup(LI2(T,S)) then
         UM:=Phi2(LI3(T,S),T,S);
      end if;
      return Scal(4.0)*Sqr(T_2)*T_2/Z_2*(Scal(One)/(Two*T_2)-Scal(One)/(Two*T_2+Z_2*C1+Z_2))+UM-C3;
   end Func2;


   function Funk2(T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin2I(RI2(T,S),T,S);
      T_2: constant Scalar:=S2(T,S);
      Z_2: constant Scalar:=X2(T,S);
      C2: Scalar:=Phi1(Scal(Zero));
   begin
      if Sup(RI3(T,S))<Inf(RI2(T,S)) then
         UM:=Phi2(RI3(T,S),T,S);
      end if;
      return Scal(4.0)*Sqr(T_2)*T_2/Z_2*(Scal(One)/(Two*T_2)-Scal(One)/(Two*T_2+Z_2*C2-Z_2))+UM-C4;
   end Funk2;

   function Func3(T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin3I(LI3(T,S),T,S);
      T_3: constant Scalar:=T3(T,S);
      Z_3: constant Scalar:=Z3(T,S);
      C1: Scalar:=Phi1(-Scal(Half));
   begin
      if Inf(LI2(T,S))>Sup(LI3(T,S)) then
         UM:=Phi3(LI2(T,S),T,S);
      end if;
      return Scal(4.0)*Sqr(T_3)*T_3/Z_3*(Scal(One)/(Two*T_3)-Scal(One)/(Two*T_3+Z_3*C1+Z_3))+UM-C3;
   end Func3;

   function Funk3(T,S: in Scalar) return Scalar is
      UM: Scalar:=Lin3I(RI3(T,S),T,S);
      T_3: constant Scalar:=S3(T,S);
      Z_3: constant Scalar:=X3(T,S);
      C2: Scalar:=Phi1(-Scal(Zero));
   begin
      if Sup(RI2(T,S))<Inf(RI3(T,S)) then
         UM:=Phi3(RI2(T,S),T,S);
      end if;
      return Scal(4.0)*Sqr(T_3)*T_3/Z_3*(Scal(One)/(Two*T_3)-Scal(One)/(Two*T_3+Z_3*C2-Z_3))+UM-C4;
   end Funk3;

   function Cee2(T: in Scalar) return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-10);
      De: Scalar;
      S: Scalar:=Es;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center((Funk2(T,S+Scal(1.0E-8))-Funk2(T,S))/Scal(1.0E-8));
         Delt:=Center(Funk2(T,S)/De);
         S:=Center(S-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return S;
   end Cee2;

   function Cee3(T: in Scalar) return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-10);
      De: Scalar;
      S: Scalar:=Es;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center((Funk3(T,S+Scal(1.0E-8))-Funk3(T,S))/Scal(1.0E-8));
         Delt:=Center(Funk3(T,S)/De);
         S:=Center(S-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return S;
   end Cee3;

   function Zed2(T: in Scalar) return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-10);
      De: Scalar;
      S: Scalar:=Es;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center((Func2(T,S+Scal(1.0E-8))-Func2(T,S))/Scal(1.0E-8));
         Delt:=Center(Func2(T,S)/De);
         S:=Center(S-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return S;
   end Zed2;

   function Zed3(T: in Scalar) return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-10);
      De: Scalar;
      S: Scalar:=Es;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center((Func3(T,S+Scal(1.0E-8))-Func3(T,S))/Scal(1.0E-8));
         Delt:=Center(Func3(T,S)/De);
         S:=Center(S-Delt);
         Err:=Sup(abs(Delt));
      end loop;

      return S;
   end Zed3;


   function Corner_23 return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-16);
      De: Scalar;
      T: Scalar:=Te;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center( ( (Cee2(T+Scal(1.0E-8))-Cee3(T+Scal(1.0E-8))) - (Cee2(T)-Cee3(T)) )/Scal(1.0E-8)) ;
         Delt:=Center( (Cee2(T)-Cee3(T))/De );
         T:=Center(T-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return T;
   end Corner_23;


   function Corner23 return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-16);
      De: Scalar;
      T: Scalar:=Te;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center( ( (Zed2(T+Scal(1.0E-8))-Zed3(T+Scal(1.0E-8))) - (Zed2(T)-Zed3(T)) )/Scal(1.0E-8)) ;
         Delt:=Center( (Zed2(T)-Zed3(T))/De );
         T:=Center(T-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return T;
   end Corner23;

   function Corner33 return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-8);
      De: Scalar;
      T: Scalar:=Te;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
      TDelt: Scalar:=Scal(1.0E-8);
   begin
      while Err>Eps loop
         De:=Center( ( (Zed2(T+TDelt)-Cee2(T+TDelt)) - (Zed2(T)-Cee2(T)) )/TDelt) ;
         Delt:=Center( (Zed2(T)-Cee2(T))/De );
         T:=Center(T-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return T;
   end Corner33;

   function Corner13(S: Scalar) return Scalar is
      use Reps.IO;
      use Ada.Text_IO;
      Eps: constant Rep:=Rep(1.0E-16);
      De: Scalar;
      T: Scalar:=Te;
      Err: Rep:=Rep(10.0);
      Delt: Scalar;
   begin
      while Err>Eps loop
         De:=Center( ( (S-Zed2(T+Scal(1.0E-8))) - (S-Zed2(T)) )/Scal(1.0E-8)) ;
         Delt:=Center( (S-Zed2(T))/De );
         T:=Center(T-Delt);
         Err:=Sup(abs(Delt));
      end loop;
      return T;
   end Corner13;

   -----------------------------------------------

   function NewT(Y,X,T,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
   begin
      return -Alpha(Y,X,Eps)*T*S/(Two*X*Beta(Y,X,Eps));
   end NewT;

   function D2(Z,W,T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(Scal(-Half));
      C2: Scalar:=Phi1(Scal(Zero));
      Delt: Scalar:=Scal(8.0E-1);
      C,CC: Scalar;
   begin
      C:=(Upp_2(Z+Delt,T,S)- ( W*Upp_2(Z,T,S)+(Scal(One)-W)*Lou_2(Z,T,S) )   )/Delt;

      CC:=(  ( W*Upp_2(Z,T,S)+(Scal(One)-W)*Lou_2(Z,T,S) ) - Lou_2(Z-Delt,T,S)  )/Delt;


      --      C:=(W*Upp_2(Z+Delt,T,S)+(Scal(One)-W)*Lou_2(Z+Delt,T,S)-
      --            ( W*Upp_2(Z,T,S)+(Scal(One)-W)*Lou_2(Z,T,S)  )  )/Delt;
--
--      CC:=(W*Upp_2(Z,T,S)+(Scal(One)-W)*Lou_2(Z,T,S)-
--            ( W*Upp_2(Z-Delt,T,S)+(Scal(One)-W)*Lou_2(Z-Delt,T,S)  )  )/Delt;

      if Sup(CC)>Sup(C) then
         return CC;
      else
         return C;
      end if;

--      if Inf(Z)>Sup(C1) then
--       return (Scal(One)-C1)*Upp_2(Z,T,S)/((Z-C1)*(Scal(One)-Z));
--      else
--       return (Scal(One)+C1)*Upp_2(Z,T,S)/((Z-C1)*(Scal(One)+Z));
--      end if;
--      if Sup(C)>Sup(CC) then
--       return C;
--      else
--          return CC;
--      end if;
   end D2;

   function D3(Z,W,T,S: in Scalar) return Scalar is
      C1: Scalar:=Phi1(Scal(-Half));
      C2: Scalar:=Phi1(Scal(Zero));
      Delt: Scalar:=Scal(8.0E-1);
      C,CC: Scalar;
   begin
      C:=(Upp_3(Z+Delt,T,S)- ( W*Upp_3(Z,T,S)+(Scal(One)-W)*Lou_3(Z,T,S) )   )/Delt;

      CC:=(  ( W*Upp_3(Z,T,S)+(Scal(One)-W)*Lou_3(Z,T,S) ) - Lou_3(Z-Delt,T,S)  )/Delt;

      if Sup(CC)>Sup(C) then
         return CC;
      else
         return C;
      end if;


--      if Inf(Z)>Sup(C1) then
--         return (Scal(One)-C1)*Upp_3(Z,T,S)/((Z-C1)*(Scal(One)-Z));
--      else
--         return (Scal(One)+C1)*Upp_3(Z,T,S)/((Z-C1)*(Scal(One)+Z));
--      end if;
      --if Sup(C)>Sup(CC) then
      --   return C;
      --else
      --   return CC;
      --end if;
   end D3;

   function D22(Y,X,Z,Q,TT,S: in Scalar) return Scalar is
      C,CC: Scalar;
   begin
      C:=(Scal(Half)-( ( Q*Upp_2(Z,TT,S)+(Scal(One)-Q)*Lou_2(Z,TT,S) )  ))/(Phi1(T(Y,X,Scal(Zero),-Sqrt(Scal(One)-X)))-Z);
      CC:=(  ( Q*Upp_2(Z,TT,S)+(Scal(One)-Q)*Lou_2(Z,TT,S) )    -Scal(Zero))/(Z-Phi1(T(Y,X,Scal(Zero),Scal(Zero))));
      if Sup(CC)>Sup(C) then
         return CC;
      else
         return C;
      end if;
   end D22;

   function D33(Y,X,Z,Q,TT,S: in Scalar) return Scalar is
      C,CC: Scalar;
   begin
      C:=(Scal(Half)-( Q*Upp_3(Z,TT,S)+(Scal(One)-Q)*Lou_3(Z,TT,S) ) )/(Phi1(T(Y,X,Scal(Zero),-Sqrt(Scal(One)-X)))-Z);
      CC:=(( Q*Upp_3(Z,TT,S)+(Scal(One)-Q)*Lou_3(Z,TT,S) )-Scal(Zero))/(Z-Phi1(T(Y,X,Scal(Zero),Scal(Zero))));


      if Sup(CC)>Sup(C) then
         return CC;
      else
         return C;
      end if;
   end D33;

   function G2(Y,X,W,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      C: constant Scalar:=Phi1(T(Y,X,Eps,-Sqrt(Scal(One)-Sqr(X))));
   begin
      return W*Upp_2(C,TT,S)+(Scal(One)-W)*Lou_2(C,TT,S);
   end G2;

   function G3(Y,X,W,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      C: constant Scalar:=Phi1(T(Y,X,Eps,-Sqrt(Scal(One)-Sqr(X))));
   begin
      return W*Upp_3(C,TT,S)+(Scal(One)-W)*Lou_3(C,TT,S);
   end G3;

   function U(Y,X,W,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      C: constant Scalar:=T(Y,X,Eps,-Sqrt(Scal(One)-Sqr(X)));
   begin
      return Sqrt(Y-(W*Upp(C,TT,S)+(Scal(One)-W)*Lou(C,TT,S)));
   end U;

   function NewS2(Y,X,W,Q,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      S1: constant Scalar:=Sqrt(Scal(One)-Sqr(X));
      S2: constant Scalar:=T(Y,X,Eps,-S1);
      S3: constant Scalar:=Phi1(S2);
      Z1: constant Scalar:=U(Y,X,W,TT,S);
      Z2: constant Scalar:=T(Y,X,Eps,Z1);
      Z3: constant Scalar:=Phi1(Z2);
      --Z4: constant Scalar:=Lou_2(Z3,TT,S);
      Z4: constant Scalar:=W*Upp_2(Z3,TT,S)+(Scal(One)-W)*Lou_2(Z3,TT,S);
   begin
      return
        -Alpha(Y,X,Eps)*
        TI(Y,X,Eps,Scal(Zero))*
        DerPhi1(S2)*
        DerTheta2(G2(Y,X,Q,TT,S,Eps),TT,S)*
        D22(Y,X,S3,Q,TT,S)*
--        D2(S3,Q,TT,S)*
        DerPhi1(Z2)*
        DerTheta2(Z4,TT,S)*
        D2(Z3,W,TT,S) /
        (Two*Z1*S1);

      --Alpha(Y,X,Eps)*
        --TI(Y,X,Eps,Scal(Zero))*
        --DerPhi1(S2)*
        --DerTheta2(G2(Y,X,W,TT,S,Eps),TT,S)*
        --D2(S3,TT,S)*
        --DerPhi1(Z2)*
        --DerTheta2(Z4,TT,S)*
        --D2(Z3,TT,S) /
        --(Two*Z1*S1);
   end NewS2;

   function NewS3(Y,X,W,Q,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      S1: constant Scalar:=Sqrt(Scal(One)-Sqr(X));
      S2: constant Scalar:=T(Y,X,Eps,-S1);
      S3: constant Scalar:=Phi1(S2);
      Z1: constant Scalar:=U(Y,X,W,TT,S);
      Z2: constant Scalar:=T(Y,X,Eps,Z1);
      Z3: constant Scalar:=Phi1(Z2);
      --      Z4: constant Scalar:=Lou_3(Z3,TT,S);
      Z4: constant Scalar:=W*Upp_3(Z3,TT,S)+(Scal(One)-W)*Lou_3(Z3,TT,S);
   begin
      return
        -        Alpha(Y,X,Eps)*
        TI(Y,X,Eps,Scal(Zero))*
        DerPhi1(S2)*
        DerTheta3(G3(Y,X,Q,TT,S,Eps),TT,S)*
        D33(Y,X,S3,Q,TT,S)*
        --D3(S3,Q,TT,S)*
        DerPhi1(Z2)*
        DerTheta3(Z4,TT,S)*
        D3(Z3,W,TT,S) /
        (Two*Z1*S1);

        --Alpha(Y,X,Eps)*
        --TI(Y,X,Eps,Scal(Zero))*
        --DerPhi1(S2)*
        --DerTheta3(G3(Y,X,W,TT,S,Eps),TT,S)*
        --D3(S3,TT,S)*
        --DerPhi1(Z2)*
        --DerTheta3(Z4,TT,S)*
        --D3(Z3,TT,S) /
        --(Two*Z1*S1);
   end NewS3;

   function NewSS(Y,X,W,Q,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      C,CC: Scalar;
   begin
      C:= NewS2(Y,X,W,Q,TT,S,Eps);
      CC:=NewS3(Y,X,W,Q,TT,S,Eps);
      if Sup(C)<Sup(CC) then
         return C;
      else
         return CC;
      end if;
   end NewSS;

   function NewS(N: in Integer; Y,X,TT,S: in Scalar; Eps: in Scalar:=Scal(Zero)) return Scalar is
      C,CC,W,Q: Scalar;
      WDelt: Scalar;
      QDelt: Scalar;
   begin
      WDelt:=Scal(Sup(Scal(One)/Scal(N)));
      QDelt:=Scal(Sup(Scal(One)/Scal(N)));
      W:=Scal(0.0);
      Q:=Scal(0.0);
      C:=NewSS(Y,X,W,Q,TT,S,Eps);
      for I in 1..N loop
         Q:=Scal(0.0);
         for J in 1..N loop
            -- W:=Scal(Zero);
            -- Q:=Scal(Zero);
            CC:=NewSS(Y,X,W,Q,TT,S,Eps);
            if Sup(CC)>Sup(C) then
               C:=CC;
            end if;
            Q:=Scal(Sup(Q),Sup(Q+QDelt));
         end loop;
         W:=Scal(Sup(W),Sup(W+WDelt));
      end loop;
      return C;
   end NewS;

   -----------------------------------------------

   function L_P(TT,S: in Scalar; Epsi: in Scalar:=Scal(Zero)) return Scalar is
      XX,XXX,Z: Scalar;
      B: constant Scalar:=Scal(One);
      Eps: constant Rep:=Rep(1.0E-7);
      Err: Scalar:=Scal(100.0);
   begin
      XX:=Scal(-0.0001);
      XXX:=Scal(-0.9999);
      while (Sup(abs(Err))>Eps or Inf(Err)<Zero) loop
         Z:=Center(Half*(XX+XXX));
         Err:=Z-Upp(T(B,Z,Epsi,Sqrt(B-Upp(T(B,Z,Epsi,-Sqrt(B-Sqr(Z))),TT,S))),TT,S);
         if Inf(Err)>Zero then XX:=Z; else XXX:=Z; end if;
      end loop;
      return Z;
   end L_P;

   function L_M(TT,S: in Scalar; Epsi: in Scalar:=Scal(Zero)) return Scalar is
      XX,XXX,Z: Scalar;
      B: constant Scalar:=Scal(One);
      Eps: constant Rep:=Rep(1.0E-7);
      Err: Scalar:=Scal(100.0);
   begin
      XXX:=Scal(-0.00000001);
      XX:=Scal(-0.99999999);
      while (Sup(abs(Err))>Eps or Inf(Err)<Zero) loop
         Z:=Center(Half*(XX+XXX));
         Err:=Lou(T(B,Z,Epsi,Sqrt(B-Lou(T(B,Z,Epsi,-Sqrt(B-Sqr(Z))),TT,S))),TT,S)-Z;
         if Inf(Err)>Zero then XX:=Z; else XXX:=Z; end if;
      end loop;
      return Z;
   end L_M;

   ------------------------------------------------

   procedure W(Theta,Y,X: in Scalar; S1,S2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
      RR: Scalar;
      R: Scalar:=Half*(R1-L1);
      Cen1: Scalar:=Half*(R1+L1);
      Theta1: Scalar:=Half*(ScalPi-ScalPi*Kappa1);
      Y1: Scalar:=R*Sin(Theta1)/Cos(Theta1);
      R1: Scalar:=Sqrt( Sqr(R)+Sqr(Y1) );

      Z1: Scalar:= Cen1    +  Cos(Theta)*R1;
      Z2: Scalar:=-abs(Y1) +  Sin(Theta)*R1;

   begin
      --(-x/alpha-beta-i*y/alpha)^2=
      --(-x/alpha-beta)^2+
      -- i*2*(x/alpha+beta)*y/alpha
      -- -y^2/alpha^2
      S1:=Y-X*(Y-( Sqr(TI(Y,X,Eps,Z1))-Sqr(Z2/Alpha(Y,X,Eps)) ) ) ;
      S2:=-X*(-Two*TI(Y,X,Eps,Z1)*(-Z2)/(Alpha(Y,X,Eps)));
      R:= Sqrt(Sqr(S1)+Sqr(S2));
      RR:=Scal(Rep'Max(Zero,Inf(R-S1)),Sup(R-S1));
      if Inf(S2)>=Zero then
         S2:=Sqrt(RR)/Sqrt(Scal(Two));
      elsif Sup(S2)<Zero then
         S2:=-Sqrt(RR)/Sqrt(Scal(Two));
      else
         S2:=Scal(
                  Inf(-Sqrt(RR)/Sqrt(Scal(Two))),
                  Sup(Sqrt(RR)/Sqrt(Scal(Two)))
                 );
      end if;
      S1:=Sqrt(R+S1)/Sqrt(Scal(Two));
      --------------------------------
      S1:= T(Y,X,Eps,-S1);
      S2:= Alpha(Y,X,Eps)*S2;
   end W;

   procedure PhiW(Theta,Y,X: in Scalar; S1,S2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
   begin
--      Put("ere1");
      W(Theta,Y,X,S1,S2,Eps);
--      Put("ere2");
      Phi1(S1,S2);
--      Put("ere3");
   end PhiW;

   function ArcCos(Z: in Scalar) return Scalar is
      D,TDelta: Scalar;
      Theta: Scalar:=Scal(Zero);
      Err: Rep:=Rep(10.0);
      Eps: Rep:=Rep(1.0E-12);
   begin
      --Sin(Z)/Cos(Z)=Z
      Theta:=Rep(0.5)*ScalPi;

      while Err>Eps loop
         D:=-Sin(Theta);
         TDelta:=-Center((Cos(Theta)-Z)/D);
         Theta:=Theta+TDelta;
         Err:=Sup(abs(TDelta));
      end loop;

      while Sup(Theta)>Sup(ScalPi) loop
         Theta:=Theta-ScalPi;
      end loop;

      while Inf(Theta)<Inf(-ScalPi) loop
         Theta:=Theta+ScalPi;
      end loop;

      return Theta;
   end ArcCos;

   procedure P(Theta,Y,X: in Scalar; A1,B1,Y1,A2,B2,Y2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
      Z1,Z2,K,M,S2: Scalar;
      Theta1: Scalar:=Half*(ScalPi-ScalPi*Kappa1);
   begin
      --      A1:=Scal(-One+Rep(1.0E-10));
      A1:=Scal(0.83);
      B2:=Scal(One-Rep(1.0E-4));
      -- Put("here1");

      PhiW(Half*ScalPi,Y,X,B1,S2,Eps);
      --Put("here2");
      --PhiW( Scal(Sup(Theta1)),Y,X,A2,S2,Eps);
      --Put("here3");

      A2:=Scal(0.93);

      PhiW(Theta,Y,X,Z1,Z2,Eps);

      --Put("here4");
      A2:=Scal(Rep'Max( Inf(Z1-
                              Scal(Rep'Max( Sup(B2-Z1),Sup(Two*abs(Z2)-(B2-Z1)  ) ) )
                           ),
                        Inf(A2)));
      K:=Sqr(Z1-Half*(A1+B1));
      M:=Scal(0.25)*Sqr(B1-A1);
      if Sup(Z2)>Zero and Inf(Z2)<Zero then
         Z2:=Scal(Sup(Z2));
         A1:=B1-Z2;
      end if;
      Y1:=Half*((K-M)/Z2+Z2);

      K:=Sqr(Z1-Half*(A2+B2));
      M:=Scal(0.25)*Sqr(B2-A2);
      if Sup(Z2)>Zero and Inf(Z2)<Zero then
         Z2:=Scal(Sup(Z2));
         A2:=Scal(0.97);
      end if;
      Y2:=Half*((K-M)/Z2+Z2);
   end P;

   procedure PP(A1,B1,Z1,Z2: in  Scalar; Y: out Scalar) is
      K,M: Scalar;
   begin
      K:=Sqr(Z1-Half*(A1+B1));
      M:=Scal(0.25)*Sqr(B1-A1);
      if Sup(Z2)>Zero and Inf(Z2)<Zero then
         Y:=Half*((K-M)/Scal(Sup(Z2))+Z2);
      else
         Y:=Half*((K-M)/Z2+Z2);
      end if;
   end PP;

   procedure H2(A,B,Y: in  out Scalar; Lamb,T,S: in Scalar) is
      D: Scalar;
   begin
      D:=(B-A);
      A:=Lou_2(A,T,S);
      B:=Upp_2(B,T,S);
      Y:=Y*(B-A)/D;
   end H2;


   procedure H3(A,B,Y: in  out Scalar; Lamb,T,S: in Scalar) is
      D: Scalar;
   begin
      D:=(B-A);
      A:=Lou_3(A,T,S);
      B:=Upp_3(B,T,S);
      Y:=Y*(B-A)/D;
   end H3;


   procedure U2(Theta,Y,X,T,S: in Scalar; A1,B1,Y1,A2,B2,Y2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
      D1,D2: Scalar;
   begin
      P(Theta,Y,X,A1,B1,Y1,A2,B2,Y2,Eps);
--      New_Line;
--      Put("Y1=");Show(Y1);
--      Put("Y2=");Show(Y2);
      D1:=(B1-A1);
      D2:=(B2-A2);

      A1:=Lou_2(A1,T,S);
      A2:=Lou_2(A2,T,S);
      B1:=Upp_2(B1,T,S);
      B2:=Upp_2(B2,T,S);
      Y1:=Y1*abs(B1-A1)/abs(D1);
      Y2:=Y2*abs(B2-A2)/abs(D2);
--      Put("New Y1=");Show(Y1);
--      Put("New Y2=");Show(Y2);
   end U2;

   procedure U3(Theta,Y,X,T,S: in Scalar; A1,B1,Y1,A2,B2,Y2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
      D1,D2: Scalar;
   begin
      P(Theta,Y,X,A1,B1,Y1,A2,B2,Y2,Eps);
      D1:=(B1-A1);
      D2:=(B2-A2);

      A1:=Lou_3(A1,T,S);
      A2:=Lou_3(A2,T,S);
      B1:=Upp_3(B1,T,S);
      B2:=Upp_3(B2,T,S);
      Y1:=Y1*abs(B1-A1)/abs(D1);
      Y2:=Y2*abs(B2-A2)/abs(D2);
   end U3;

   procedure V2(Theta,Y,X,T,S: in Scalar; A1,B1,Y1,A2,B2,Y2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
   begin
      U2(Theta,Y,X,T,S,A1,B1,Y1,A2,B2,Y2,Eps);
   end V2;

   procedure V3(Theta,Y,X,T,S: in Scalar; A1,B1,Y1,A2,B2,Y2: out Scalar; Eps: in Scalar:=Scal(Zero)) is
   begin
      U3(Theta,Y,X,T,S,A1,B1,Y1,A2,B2,Y2,Eps);
   end V3;

   procedure M(Y: in Scalar; Z1,Z2: in out Scalar) is
      R,R1,R2: Scalar;
   begin

      Z1:=Y-Z1;
      R:=Sqrt(Sqr(Z1)+Sqr(Z2));
      R1:=Scal(Rep'Max(Zero,Inf(R-Z1)),Sup(R-Z1));
      R2:=Scal(Rep'Max(Zero,Inf(R+Z1)),Sup(R+Z1));



      Z2:=Sqrt(R1)/Sqrt(Scal(Two));--im part
      Z1:=Sqrt(R2)/Sqrt(Scal(Two));--real par
   end M;

   procedure Nbound(Y,X: in Scalar; Z1,Z2: in out Scalar; Eps: in Scalar:=Scal(Zero)) is
   begin
      M(Y,Z1,Z2);
      Z1:=T(Y,X,Eps,-Z1);
      Z2:=Alpha(Y,X,Eps)*Z2;
      Phi1(Z1,Z2);
   end Nbound;

end Doubling;

