with Reps.Ops, Ada.Numerics.Generic_Elementary_Functions;

pragma Elaborate_All (Ada.Numerics.Generic_Elementary_Functions);
pragma Elaborate_All (Reps.Ops);

package body Numerics is

  package Numeric_EF is new Ada.Numerics.Generic_Elementary_Functions(Numeric);

  function IsNumeric(S: Numeric) return Boolean is
  begin
    return True;
  end IsNumeric;

  function Scal(K: Integer) return Numeric is
  begin
    return Numeric(K);
  end Scal;

  function Scal(R: Rep) return Numeric is
  begin
    return Numeric(R);
  end Scal;

  function Approx(S: Numeric) return Rep is
  begin
    return Rep(S);
  end Approx;

  function Ball(R: Rep) return Numeric is
  begin
    return 0.0;
  end Ball;

  function Ball(S: Numeric) return Numeric is
  begin
    return 0.0;
  end Ball;

  function Scal(R1,R2: Rep) return Numeric is
  begin
    return Numeric(Half*(R1+R2));
  end Scal;

  function Center(S: Numeric) return Numeric is
  begin
     return S;
  end Center;

  function Contains(S: Numeric; Z: Rep) return Boolean is
  begin
     return S=Numeric(Z);
  end Contains;

  procedure ResetCenter(S: in out Numeric) is
  begin
     S := 0.0;
  end Resetcenter;

  procedure Split(S: in Numeric; S0,SE: out Numeric) is
  begin
    S0 := S;
    SE := 0.0;
  end Split;

  function Union(S1,S2: Numeric) return Numeric is
  begin
    return 0.0;
  end Union;

  function InfAbs(S: Numeric) return Rep is
  begin
    return Rep(abs(S));
  end InfAbs;

  function Inv(S: Numeric) return Numeric is
  begin
    return 1.0/S;
  end Inv;

  function "*"(R: Rep; S: Numeric) return Numeric is
  begin
    return Numeric(R)*S;
  end "*";

  procedure ErrMult(R: in Rep; S: in out Numeric) is
  begin
    null;
  end ErrMult;

  function "/"(S: Numeric; R: Rep) return Numeric is
  begin
    return S/Numeric(R);
  end "/";

  function Sqr(S: Numeric) return Numeric is
  begin
    return S*S;
  end Sqr;

  procedure Ln(S1,S2: in  out Numeric) is
     use Reps.Ops;
  begin
     S2:=Numeric(ArcTan(Rep(S1/S2)));
     S1:=Log(S1);
  end Ln;

  function Cosh(S: Numeric) return Numeric renames Numeric_EF.Cosh;
  function Sinh(S: Numeric) return Numeric renames Numeric_EF.Sinh;
  function Cos(S: Numeric) return Numeric renames Numeric_EF.Cos;
  function Sin(S: Numeric) return Numeric renames Numeric_EF.Sin;
  function Exp(S: Numeric) return Numeric renames Numeric_EF.Exp;
  function Log(S: Numeric) return Numeric renames Numeric_EF.Log;
  function Sqrt(S: Numeric) return Numeric renames Numeric_EF.Sqrt;

  function TruncExp(N: Natural; S: Numeric) return Numeric is
    Y: Numeric := 1.0;
  begin
    for I in reverse 1 .. N loop
      Y := 1.0+S*Y/Numeric(I);
    end loop;
    return Y;
  end TruncExp;

  function RemExp(N: Natural; S: Numeric) return Numeric is
--- Taylor remainder for Exp, divided by S**(N+1)/(N+1)!
    M: Integer;
    Y: Numeric;
  begin
    if S<0.0 then
      raise Undefined;
    elsif S<0.5 then
      M := Integer'Max(N+1,17);
    elsif S<2.0 then
      M := Integer'Max(N+1,26);
    elsif S<8.0 then
      M := Integer'Max(N+1,46);
    elsif S<32.0 then
      M := Integer'Max(N+1,96);
    elsif S<128.0 then
      M := Integer'Max(N+1,245);
    elsif S<300.0 then
      M := Integer'Max(N+1,470);
    elsif S<512.0 then
      M := Integer'Max(N+1,730);
    else
      raise Undefined;
    end if;
    Y := Inv(1.0-S/Numeric(M+1));
    for I in reverse N+2 .. M loop
      Y := 1.0+(S*Y)/Numeric(I);
    end loop;
    return Y;
  end RemExp;

end Numerics;
