with Reps;
use Reps;

package Numerics is

  pragma Pure;

  type Numeric is new Rep;

--- basic
  function IsNumeric(S: Numeric) return Boolean;
  function Err0(S: Numeric) return Boolean renames IsNumeric;
  function Scal(K: Integer) return Numeric;
  function Scal(R: Rep) return Numeric;
  function Approx(S: Numeric) return Rep;

  pragma Inline (IsNumeric,Err0,Scal,Approx);

--- sets
  function Ball(R: Rep) return Numeric;
  function Ball(S: Numeric) return Numeric;
  function Scal(R1,R2: Rep) return Numeric;
  function Center(S: Numeric) return Numeric;
  function Contains(S: Numeric; Z: Rep) return Boolean;
  procedure ResetCenter(S: in out Numeric);
  function ResetCenter(S: Numeric) return Numeric renames Ball;
  procedure Split(S: in Numeric; S0,SE: out Numeric);
  function Widen(S: Numeric) return Numeric renames Center;
  function Union(S1,S2: Numeric) return Numeric;

  pragma Inline (Ball,Center,ResetCenter,Split);

--- order
  function Inf(S: Numeric) return Rep renames Approx;
  function Sup(S: Numeric) return Rep renames Approx;
  function Min(S1,S2: Numeric) return Numeric renames Numeric'Min;
  function Max(S1,S2: Numeric) return Numeric renames Numeric'Max;

  pragma Inline (Inf,Sup);

--- ops
  function Norm(S: Numeric) return Numeric renames "abs";
  function InfAbs(S: Numeric) return Rep;
  function SupAbs(S: Numeric) return Rep renames InfAbs;
  function "*"(R: Rep; S: Numeric) return Numeric;
  procedure ErrMult(R: in Rep; S: in out Numeric);
  function "/"(S: Numeric; R: Rep) return Numeric;
  function Sqr(S: Numeric) return Numeric;
  function Inv(S: Numeric) return Numeric;

  pragma Inline (InfAbs,SupAbs,ErrMult,Norm,Sqr,Inv);

--- fun
  function Cosh(S: Numeric) return Numeric;
  function Sinh(S: Numeric) return Numeric;
  function Cos(S: Numeric) return Numeric;
  function Sin(S: Numeric) return Numeric;
  function Exp(S: Numeric) return Numeric;
  function Log(S: Numeric) return Numeric;
  procedure Ln(S1,S2: in out Numeric);
  function Sqrt(S: Numeric) return Numeric;
  function TruncExp(N: Natural; S: Numeric) return Numeric;
  function RemExp(N: Natural; S: Numeric) return Numeric;

end Numerics;
