with Reps, Numerics;
use Reps, Numerics;

package S_Numerics is

  pragma Pure;

  subtype Numeric is Numerics.Numeric;

--- basic
  function S_IsNumeric(S: Numeric) return Boolean renames IsNumeric;
  function S_Err0(S: Numeric) return Boolean renames Err0;
  function S_Scal(K: Integer) return Numeric renames Scal;
  function S_Scal(R: Rep) return Numeric renames Scal;
  function S_Approx(S: Numeric) return Rep renames Approx;

--- sets
  function S_Ball(R: Rep) return Numeric renames Ball;
  function S_Ball(S: Numeric) return Numeric renames Ball;
  function S_Scal(R1,R2: Rep) return Numeric renames Scal;
  function S_Center(S: Numeric) return Numeric renames Center;
  function S_Contains(S: Numeric; Z: Rep) return Boolean renames Contains;
  procedure S_ResetCenter(S: in out Numeric) renames ResetCenter;
  function S_ResetCenter(S: Numeric) return Numeric renames ResetCenter;
  procedure S_Split(S: in Numeric; S0,SE: out Numeric) renames Split;
  function S_Widen(S: Numeric) return Numeric renames Widen;
  function S_Union(S1,S2: Numeric) return Numeric renames Union;

--- order
  function S_Inf(S: Numeric) return Rep renames Inf;
  function S_Sup(S: Numeric) return Rep renames Sup;
  function S_LT(S1,S2: Numeric) return Boolean renames "<";
  function S_LE(S1,S2: Numeric) return Boolean renames "<=";
  function S_GT(S1,S2: Numeric) return Boolean renames ">";
  function S_GE(S1,S2: Numeric) return Boolean renames ">=";
  function S_Min(S1,S2: Numeric) return Numeric renames Min;
  function S_Max(S1,S2: Numeric) return Numeric renames Max;

--- ops
  function S_Norm(S: Numeric) return Numeric renames Norm;
  function S_InfAbs(S: Numeric) return Rep renames InfAbs;
  function S_SupAbs(S: Numeric) return Rep renames SupAbs;
  function S_Neg(S: Numeric) return Numeric renames "-";
  function S_Sum(S,T: Numeric) return Numeric renames "+";
  function S_Diff(S,T: Numeric) return Numeric renames "-";
  function S_Prod(R: Rep; S: Numeric) return Numeric renames "*";
  procedure S_ErrMult(R: in Rep; S: in out Numeric) renames ErrMult;
  function S_Quot(S: Numeric; R: Rep) return Numeric renames "/";
  function S_Sqr(S: Numeric) return Numeric renames Sqr;
  function S_Prod(S,T: Numeric) return Numeric renames "*";
  function S_Inv(S: Numeric) return Numeric renames Inv;
  function S_Quot(S,T: Numeric) return Numeric renames "/";
  function S_Pow(S: Numeric; I: Integer) return Numeric renames "**";

--- fun
  function S_Cosh(S: Numeric) return Numeric renames Cosh;
  function S_Sinh(S: Numeric) return Numeric renames Sinh;
  function S_Cos(S: Numeric) return Numeric renames Cos;
  function S_Sin(S: Numeric) return Numeric renames Sin;
  function S_Exp(S: Numeric) return Numeric renames Exp;
  function S_Log(S: Numeric) return Numeric renames Log;
  function S_Sqrt(S: Numeric) return Numeric renames Sqrt;
  function S_TruncExp(N: Natural; S: Numeric) return Numeric renames TruncExp;
  function S_RemExp(N: Natural; S: Numeric) return Numeric renames RemExp;

end S_Numerics;
