with Messages;
use Messages;

package body Taylors is

  Zero_Taylor: Taylor;

  function IMin(I1,I2: Integer) return Integer renames Integer'Min;
  function IMax(I1,I2: Integer) return Integer renames Integer'Max;

  function Rho(F: Taylor) return Radius is
  begin
    return F.R;
  end Rho;

  procedure SetRho(R: in Radius; F: out Taylor) is
  begin
    F.R := R;
  end SetRho;

  function IsZero(F: Taylor) return Boolean is
  begin
    if not IsZero(F.C) then return False; end if;
    if Trunc then return True; end if;
    return F.Eh=ScalZero and then F.Eg=ScalZero;
  end IsZero;

  procedure SetZero(R: in Radius; F: out Taylor) is
  begin
     F := Zero_Taylor;
     F.R := R;
  end SetZero;

  procedure ResetErrComp(F: in out Taylor) is
  begin
     F.Eh := ScalZero;
     F.Eg := ScalZero;
  end ResetErrComp;


--  function EffDeg(P: Taylor) return Degree is --fine
--  begin
--     for I in reverse 0 .. PDeg loop
--        for J in reverse 0 .. PDeg loop
--           if P.C(I,J) /= ScalZero then
--              return K;
--           end if;
--        end loop;
--     end loop;
--     if Trunc or else P.E=Zero then return -1; end if;
--     return 0;
--  end EffDeg;

  function ZeroFun(R: Radius) return Taylor is
     F: Taylor := Zero_Taylor;
  begin
    F.R := R;
    return F;
  end ZeroFun;

  function Monom(R: Radius; I,J: Natural; S: Scalar) return Taylor is
    F: Taylor := Zero_Taylor;
  begin
    F.R := R;
    F.C(I,J) := S;
    return F;
  end Monom;

  function Component(I,J: Natural; F: Taylor) return Scalar is
  begin
    return F.C(I,J);
  end Component;

  function Component(F,H: Taylor) return Scalar is
     S: Scalar:=Scal(Zero);
     C: Scalar:=Scal(Zero);
  begin
     for I in Power loop
        for J in Power loop
           S:=S+F.C(I,J)*H.C(I,J);
           C:=C+Sqr(H.C(I,J));
        end loop;
     end loop;
     return S/Sqrt(C);
  end Component;

  procedure Neg(F: in out Taylor) is
  begin
    Neg(F.C);
  end Neg;

  function "-"(F: Taylor) return Taylor is
  begin
    return (F.R,-F.C,F.Eh,F.Eg);
  end "-";

  procedure Mult(S: in Scalar; F: in out Taylor) is
  begin
    Mult(S,F.C);
    F.Eh := S*F.Eh;
    F.Eg := S*F.Eg;
  end Mult;

  function "*"(S: Scalar; F: Taylor) return Taylor is
  begin
    return (F.R,S*F.C,S*F.Eh,S*F.Eg);
  end "*";

  procedure Add(F1: in Taylor; F2: in out Taylor) is
  begin
    F2.R := Radius'Min(F1.R,F2.R);
    Add(F1.C,F2.C);
    F2.Eh := F1.Eh+F2.Eh;
    F2.Eg := F1.Eg+F2.Eg;
  end Add;

  function "+"(F1,F2: Taylor) return Taylor is
    F3: Taylor;
  begin
    F3.R := Radius'Min(F1.R,F2.R);
    F3.C := F1.C+F2.C;
    F3.Eh := F1.Eh+F2.Eh;
    F3.Eg := F1.Eg+F2.Eg;
    return F3;
  end "+";

  procedure Sub(F1: in Taylor; F2: in out Taylor) is
  begin
    F2.R := Radius'Min(F1.R,F2.R);
    Sub(F1.C,F2.C);
    F2.Eh := F1.Eh+F2.Eh;
    F2.Eg := F1.Eg+F2.Eg;
  end Sub;

  function "-"(F1,F2: Taylor) return Taylor is
    F3: Taylor;
  begin
    F3.R := Radius'Min(F1.R,F2.R);
    F3.C := F1.C-F2.C;
    F3.Eh := F1.Eh+F2.Eh;
    F3.Eg := F1.Eg+F2.Eg;
    return F3;
  end "-";

  procedure Mult(F1: in Taylor; F2: in out Taylor) is
    S: Scalar;
  begin
    if IsZero(F2) then return; end if;
    F2.R := Radius'Min(F1.R,F2.R);
    for I in reverse Power loop
      for J in reverse Power loop
        S := ScalZero;
        for I1 in 0 .. I loop
          for J1 in 0 .. J loop
            S := S+F1.C(I1,J1)*F2.C(I-I1,J-J1);
          end loop;
        end loop;
        F2.C(I,J) := S;
      end loop;
    end loop;
--- error estimates ...
  end Mult;

  function "*"(F1,F2: Taylor) return Taylor is
    F3: Taylor := F2;
  begin
    Mult(F1,F3);
    return F3;
  end "*";

  procedure MultAdd(F1,F2: in Taylor; F3: in out Taylor) is
    S: Scalar;
  begin
    F3.R := Radius'Min(Radius'Min(F1.R,F2.R),F3.R);
    for I in reverse Power loop
      for J in reverse Power loop
        S := F3.C(I,J);
        for I1 in 0 .. I loop
          for J1 in 0 .. J loop
            S := S+F1.C(I1,J1)*F2.C(I-I1,J-J1);
          end loop;
        end loop;
        F3.C(I,J) := F3.C(I,J)+S;
      end loop;
    end loop;
--- error estimates ...
  end MultAdd;

  function Inv(F: Taylor) return Taylor is
     C: constant Scalar := Inv(F.C(0,0));
     Ft: Taylor := (-C)*F;
     G: Taylor;
  begin
     SetZero(Rho(F),G);
     Ft.C(0,0) := ScalZero;
     G.C(0,0) := ScalOne;
     for I in reverse 0 .. 100 loop
        Mult(Ft,G);
        G.C(0,0) := G.C(0,0)+ScalOne;
     end loop;
     G.Eh := ScalZero;
     G.Eg := ScalZero;
     --- error estimates ...
     return C*G;
  end Inv;

  function "/"(F1,F2: Taylor) return Taylor is
  begin
    return F1*Inv(F2);
  end "/";


--  procedure Quot(P1,P2: in Taylor; P3: out Taylor) is
--     R,S: Scalar;
--  begin
--     S := P2.C(0);
--     if S =Scal(Zero) then
--        Message("Taylor_Ops.Quot",Division_By_Zero);
--        return;
--     end if;
--     P3.C(0) := P1.C(0)/S;
--     for K in Power loop
--        R := P1.C(K);
--        for I in 1 .. IMin(K,Power) loop
--           R := R-P2.C(I)*P3.C(K-I);
--       end loop;
--        P3.C(K) := R/S;
--     end loop;
--  end Quot;


--  function Quot(P1,P2: Taylor; D: in Degree := Size) return Taylor is
--     Pt: Taylor;
--  begin
--     Quot(P1,P2,Pt,D);
--     return Pt;
--  end Quot;


--  function "/"(P1,P2: Taylor) return Taylor is
--     Pt: Taylor;
--  begin
--     Quot(P1,P2,Pt);
--     return Pt;
--   end "/";



  function Val(S1,S2: Scalar; F: Taylor) return Scalar is
    T: Scalar := Val(S1,S2,F.C);
  begin
     if not Trunc then
        if (SupAbs(S1)>F.R) or (SupAbs(S1)>F.R) then
           Message("Taylors.Val",Domain_Error);
        end if;
        --- error estimates ...
        T:=T+Ball(Max(Abs(((S1/F.R)**(CutOff+1))),Abs(((S2/F.R)**(CutOff+1))))*F.Eh)+F.Eg;
     end if;
     return T;
  end Val;

  function Val(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
    T: Scalar := Val(S1,S2,F.C);
  begin
     T:=(C+E)*(S1-Scal(Half))+E*(S2+Scal(Half))+Half*B*Sqr(S2+Scal(Half))+T;
     if not Trunc then
        if (SupAbs(S1)>F.R) or (SupAbs(S1)>F.R) then
           Message("Taylors.Val",Domain_Error);
        end if;
        --- error estimates ...
        T:=T+Ball(Max(Abs(((S1/F.R)**(CutOff+1))),Abs(((S2/F.R)**(CutOff+1))))*F.Eh)+F.Eg;
     end if;
     return T;
  end Val;


  function DerVal1(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
     R1: Rep:=Rep(1.001)*Rep'Max(SupAbs(S1),SupAbs(S2));
     F1: Taylor:=Der1(R1,F);
     T: Scalar :=Val(S1,S2,F1);
  begin
     return (C+E)+T;
  end DerVal1;

  function DerVal2(B,C,E,S1,S2: Scalar; F: Taylor) return Scalar is
     R1: Rep:=Rep(1.001)*Rep'Max(SupAbs(S1),SupAbs(S2));
     F1: Taylor:=Der2(R1,F);
     T: Scalar :=Val(S1,S2,F1);
  begin
     return E+B*(S2+Scal(Half))+T;
  end DerVal2;


  function Norm(R: Radius; F: Taylor) return Scalar is
     S1,S2: Scalar;
  begin
     S1 := ScalZero;
     for I in reverse Power loop
        S2 := ScalZero;
        for J in reverse Power loop
           S2 := R*S2+abs(F.C(I,J));
        end loop;
      S1 := R*S1+S2;
     end loop;
     if not Trunc then
        if F.R<R then
           Message("Taylors.Norm",Domain_Error);
        end if;
        S1 := S1+F.Eg+F.Eh*Scal(R)**(PDeg+1);
     end if;
     return S1;
  end Norm;


  function Norm(F: Taylor) return Scalar is
  begin
     return Norm(F.R,F);
  end Norm;

  function Norm(Rx,Ry: Radius; F: Taylor) return Scalar is
     S1,S2: Scalar;
  begin
     if Sup(F.Eh)>Zero or Sup(F.Eg)>Zero then
        Message("Taylors.Norm",Not_Implemented);
     end if;
     S1 := ScalZero;
     for I in reverse Power loop
        S2 := ScalZero;
        for J in reverse Power loop
           S2 := Ry*S2+abs(F.C(I,J));
        end loop;
        S1 := Rx*S1+S2;
     end loop;
     return S1;
  end Norm;

  procedure Scale(S1,S2: in Scalar; F: in out Taylor) is
  begin
    F.R := Inf(Scal(F.R)/Rep'Max(SupAbs(S1),SupAbs(S2)));
    Scale(S1,S2,F.C);
--- error estimates ...
  end Scale;

  function Scale(S1,S2: Scalar; F: Taylor) return Taylor is
    G: Taylor;
  begin
    G.R := Inf(Scal(F.R)/Rep'Max(SupAbs(S1),SupAbs(S2)));
    Scale(S1,S2,F.C,G.C);
    G.Eh := ScalZero;
    G.Eg := ScalZero;
--- error estimates ...
    return G;
  end Scale;

  procedure Fmax1(R: in Scalar; M: out Natural; F: out Scalar) is
     X: Scalar;
  begin
     X := Inv(-Log(R));
     M := Integer(Rep'Ceiling(Sup(X)));
     F := X*Exp(-ScalOne)/R;
     if M<(CutOff+1) then
        F:=Scal(CutOff+1)*(R**CutOff);
     end if;
  end Fmax1;


  procedure Fmax2(R: in Scalar; M: out Natural; F: out Scalar) is
     X,Y: Scalar;
  begin
     X := Half*((ScalOne-ScalTwo/Log(R))+Sqrt(Scal(4.0)/Sqr(Log(R))+ScalOne) );
     M := Integer(Rep'Ceiling(Sup(X)));
     F := X*(X-ScalOne)*Exp(Log(X-ScalTwo));
     Y := ScalTwo*(R**(CutOff-2));
     if Sup(Y)>Sup(F) then
        F:=Y;
     elsif M<(CutOff+1) then
        F:=Scal((CutOff+1)*CutOff)*(R**(CutOff-1));
     end if;
  end Fmax2;


  procedure Der1(R1: in Radius; F: in Taylor; T: out Taylor) is
     R: Radius:=F.R;
     M: Integer;
     E: Scalar;
  begin
     SetZero(R1,T);
     for I in 1..PDeg loop
        for J in Power loop
           T.C(I-1,J) := Rep(I)*F.C(I,J);
        end loop;
     end loop;
     if not Trunc then
        --- error estimates ...
        Fmax1(Scal(R1)/R,M,E);
        T.C(CutOff,0):=(E*F.Eh/R)/(Scal(R1)**CutOff);
        T.Eh:=ScalZero;
        T.Eg:=ScalZero;--??
     end if;
  end Der1;


  function Der1(R1: Radius; F: Taylor) return Taylor is
     T: Taylor;
  begin
     Der1(R1,F,T);
     return T;
  end Der1;


  procedure Der2(R1: in Radius; F: in Taylor; T: out Taylor) is
     R: Radius:=F.R;
     M: Integer;
     E: Scalar;
  begin
     SetZero(R1,T);
     for I in Power loop
        for J in 1..PDeg loop
           T.C(I,J-1) := Rep(J)*F.C(I,J);
        end loop;
     end loop;
     if not Trunc then
        --- error estimates ...
        Fmax1(Scal(R1)/R,M,E);
        T.C(0,CutOff):=(E*F.Eh/R)/(Scal(R1)**CutOff);
        T.Eh:=ScalZero;
        T.Eg:=ScalZero;--??
     end if;
  end Der2;


  function Der2(R1: Radius; F: Taylor) return Taylor is
     T: Taylor;
  begin
     Der2(R1,F,T);
     return T;
  end Der2;


  procedure DerDer11(R1: in Radius; F: in Taylor; T: out Taylor) is
     R: Radius:=F.R;
     M: Integer;
     E: Scalar;
  begin
     SetZero(R1,T);
     for I in 2..PDeg loop
        for J in Power loop
           T.C(I-2,J) := Rep(I)*Rep(I-1)*F.C(I,J);
        end loop;
     end loop;
     if not Trunc then
        --- error estimates ...
        Fmax2(Scal(R1)/R,M,E);
        T.C(CutOff,0):=(E*F.Eh/Sqr(Scal(R)))/(Scal(R1)**CutOff);
        T.Eh:=ScalZero;
        T.Eg:=ScalZero;--??
     end if;
  end DerDer11;


  function DerDer11(R1: Radius; F: Taylor) return Taylor is
     T: Taylor;
  begin
     DerDer11(R1,F,T);
     return T;
  end DerDer11;


  procedure DerDer22(R1: in Radius; F: in Taylor; T: out Taylor) is
     R: Radius:=F.R;
     M: Integer;
     E: Scalar;
  begin
     SetZero(R1,T);
     for I in Power loop
        for J in 2..PDeg loop
           T.C(I,J-2) := Rep(J)*Rep(J-1)*F.C(I,J);
        end loop;
     end loop;
     if not Trunc then
        --- error estimates ...
        Fmax2(Scal(R1)/R,M,E);
        T.C(0,CutOff):=(E*F.Eh/Sqr(Scal(R)))/(Scal(R1)**CutOff);
        T.Eh:=ScalZero;
        T.Eg:=ScalZero;
     end if;
  end DerDer22;


  function DerDer22(R1: Radius; F: Taylor) return Taylor is
     T: Taylor;
  begin
     DerDer22(R1,F,T);
     return T;
  end DerDer22;


  procedure DerDer12(R1: in Radius; F: in  Taylor; T: out Taylor) is
     R: Radius:=F.R;
     M: Integer;
     E: Scalar;
  begin
     SetZero(R1,T);
     for I in 1..PDeg loop
        for J in 1..PDeg loop
           T.C(I-1,J-1) := Rep(J)*Rep(I)*F.C(I,J);
        end loop;
     end loop;
     if not Trunc then
        --- error estimates ...
        Fmax1(Scal(R1)/R,M,E);
        T.C(0,CutOff):=Half*(Sqr(E)*F.Eh/Sqr(Scal(R)))/(Scal(R1)**CutOff);
        T.C(CutOff,0):=Half*(Sqr(E)*F.Eh/Sqr(Scal(R)))/(Scal(R1)**CutOff);
        T.Eh:=ScalZero;
        T.Eg:=ScalZero;
     end if;
  end DerDer12;


  function DerDer12(R1: Radius; F: Taylor) return Taylor is
     T: Taylor;
  begin
     DerDer12(R1,F,T);
     return T;
  end DerDer12;


  function Trans1(S: Scalar; F: Taylor) return Taylor is
    G: Taylor;
  begin
    G.R := Inf(Scal(F.R)-abs(S));
    Translate1(S,F.C,G.C);
    G.Eh := ScalZero;
    G.Eg := ScalZero;
--- error estimates ...
    return G;
  end Trans1;

  function Trans2(S: Scalar; F: Taylor) return Taylor is
    G: Taylor;
  begin
    G.R := Inf(Scal(F.R)-abs(S));
    Translate2(S,F.C,G.C);
    G.Eh := ScalZero;
    G.Eg := ScalZero;
--- error estimates ...
    return G;
  end Trans2;

  procedure Trans(S1,S2: in Scalar; F: in out Taylor) is
    P: Poly2;
  begin
    F.R := Inf(Scal(F.R)-Max(abs(S1),abs(S2)));
    Translate1(S1,F.C,P);
    Translate2(S2,P,F.C);
--- error estimates ...
  end Trans;

  function Trans(S1,S2: Scalar; F: Taylor) return Taylor is
    G: Taylor := F;
  begin
    Trans(S1,S2,G);
    return G;
  end Trans;

  function Invert(R: Rep; F,InvF0: Taylor) return Taylor is
     InvF: Taylor:=InvF0;
  begin
     for I in 1..200 loop
        InvF:=InvF+(Monom(R,0,1,Scal(One))-Comp2(InvF,F) )/Comp2(InvF,Der2(R,F));
     end loop;
     return InvF;
  end Invert;

  function Comp1(F,H: Taylor) return Taylor is
    R: constant Radius := Radius'Min(F.R,H.R);
    G1,G2: Taylor;
  begin
    SetZero(R,G1);
    for I in reverse Power loop
      SetZero(R,G2);
      for J in reverse Power loop
        G2.C(0,J) := H.C(I,J);
      end loop;
      Mult(F,G1);
      Add(G2,G1);
    end loop;
    if not Trunc then
      if Sup(Norm(R,F))>H.R then
        Message("Taylors.Comp1",Domain_Error);
      end if;
--- error estimates ...
    end if;
    return G1;
  end Comp1;

  function Comp2(F,H: Taylor) return Taylor is
    R: constant Radius := Radius'Min(F.R,H.R);
    G1,G2: Taylor;
  begin
    SetZero(R,G1);
    for J in reverse Power loop
      SetZero(R,G2);
      for I in reverse Power loop
        G2.C(I,0) := H.C(I,J);
      end loop;
      Mult(F,G1);
      Add(G2,G1);
    end loop;
    if not Trunc then
      if Sup(Norm(R,F))>H.R then
        Message("Taylors.Comp2",Domain_Error);
      end if;
--- error estimates ...
    end if;
    return G1;
  end Comp2;

  function Comp(F1,F2,H: Taylor) return Taylor is
    R: constant Radius := Radius'Min(F1.R,F2.R);
    G1,G2: Taylor;
  begin
    SetZero(R,G1);
    for I in reverse Power loop
      SetZero(R,G2);
      for J in reverse Power loop
        Mult(F2,G2);
        G2.C(0,0) := G2.C(0,0)+H.C(I,J);
      end loop;
      Mult(F1,G1);
      Add(G2,G1);
    end loop;
    if not Trunc then
      if (Sup(Norm(R,F1))>H.R) or (Sup(Norm(R,F2))>H.R) then
        Message("Taylors.Comp",Domain_Error);
      end if;
--- error estimates ...
    end if;
    return G1;
  end Comp;


  function SupRad(C: Scalar; Tay: Taylor) return Radius is
     R: Rep;
  begin
     R:=SupAbs(C);
     return Rep(1.000001)*R;
  end SupRad;



begin

  Zero_Taylor.R := Radius'Last;
  SetZero(Zero_Taylor.C);
  Zero_Taylor.Eh := ScalZero;
  Zero_Taylor.Eg := ScalZero;

end Taylors;
