with Reps;
use Reps;

generic

  type Scalar is private;

  Zero_Scalar: Scalar;

  with function IsZero(S: Scalar) return Boolean is <>;
  with procedure Neg(S: in out Scalar) is <>;
  with function "-"(S: Scalar) return Scalar is <>;
  with procedure Add(S1: in Scalar; S2: in out Scalar) is <>;
  with function "+"(S,T: Scalar) return Scalar is <>;
  with procedure Sub(S1: in Scalar; S2: in out Scalar) is <>;
  with function "-"(S,T: Scalar) return Scalar is <>;
  with procedure Mult(S1: in Scalar; S2: in out Scalar) is <>;
  with function "*"(S,T: Scalar) return Scalar is <>;
  --  with function "**"(S: Scalar; I: Integer) return Scalar is <>;
  --  with function "/"(S,T: Scalar) return Scalar is <>;
  --  with function "abs"(S: Scalar) return Scalar is <>;
  with procedure MultAdd(S1,S2: in Scalar; S3: in out Scalar) is <>;
  --  with function Scal(R: Rep) return Scalar is <>;
  --  with function Max(S1,S2: Scalar) return Scalar is <>;

package Vectors is

  type Vector is array(Integer range <>) of Scalar;
  type Matrix is array(Integer range <>, Integer range <>) of Scalar;
  type Tensor is array(Integer range <>, Integer range <>, Integer range <>) of Scalar;

  function IsZero(V: Vector) return Boolean;                           -- V=0
  procedure SetZero(V: out Vector);                                    -- V := 0
  procedure Neg(V: in out Vector);                                     -- V := -V
  function "-"(V: Vector) return Vector;                               -- return -V
  procedure Add(V1: in Vector; V2: in out Vector);                     -- V2 := V2+V1
  procedure Sum(V1,V2: in Vector; V3: out Vector);                     -- V3 := V1+V2
  function "+"(V1,V2: Vector) return Vector;                           -- return V1+V2
  procedure Sub(V1: in Vector; V2: in out Vector);                     -- V2 := V2-V1
  procedure Diff(V1,V2: in Vector; V3: out Vector);                    -- V3 := V1-V2
  function "-"(V1,V2: Vector) return Vector;                           -- return V1-V2
  procedure Mult(S: in Scalar; V: in out Vector);                      -- V := S*V
  procedure Prod(S: in Scalar; V1: in Vector; V2: out Vector);         -- V2 := S*V1
  function "*"(S: Scalar; V: Vector) return Vector;                    -- return S*V
  function "*"(V1,V2: Vector) return Scalar;                    -- return V1 dot V2
  procedure MultAdd(S: in Scalar; V1: in Vector; V2: in out Vector);   -- V2 := V2+S*V1

  function IsZero(A: Matrix) return Boolean;
  procedure SetZero(A: out Matrix);
  procedure ZeroRow(I: in Integer; A: in out Matrix);
  procedure SetRow(I: in Integer; V: in Vector; A: in out Matrix);
  function GetRow(I: Integer; A: Matrix) return Vector;
  procedure ZeroColumn(J: in Integer; A: in out Matrix);
  procedure SetColumn(J: in Integer; V: in Vector; A: in out Matrix);
  function GetColumn(J: Integer; A: Matrix) return Vector;
  function Transpose(A: Matrix) return Matrix;
  procedure Neg(A: in out Matrix);
  function "-"(A: Matrix) return Matrix;
  procedure Add(A1: in Matrix; A2: in out Matrix);
  procedure Sum(A1,A2: in Matrix; A3: out Matrix);
  function "+"(A1,A2: Matrix) return Matrix;
  procedure Add(A1: in Tensor; A2: in out Tensor);
  function "+"(A1,A2: Tensor) return Tensor;
  procedure Sub(A1: in Matrix; A2: in out Matrix);
  procedure Diff(A1,A2: in Matrix; A3: out Matrix);
  function "-"(A1,A2: Matrix) return Matrix;
  procedure Mult(S: in Scalar; A: in out Matrix);
  procedure Prod(S: in Scalar; A1: in Matrix; A2: out Matrix);
  function "*"(S: Scalar; A: Matrix) return Matrix;
  procedure MultAdd(S: in Scalar; A1: in Matrix; A2: in out Matrix);

  procedure Mult(A: in Matrix; V: in out Vector);
  procedure Prod(A: in Matrix; V1: in Vector; V2: out Vector);
  function "*"(A: Matrix; V: Vector) return Vector;
  procedure MultAdd(A: in Matrix; V1: in Vector; V2: in out Vector);
  procedure Mult(A1: in Matrix; A2: in out Matrix);
  procedure Mult(A1: in Matrix; A2: in out Tensor);
  procedure Mult(A1: in out Tensor; A2,A3: in Matrix);
  procedure Prod(A1,A2: in Matrix; A3: out Matrix);
  function "*"(A1,A2: Matrix) return Matrix;
  function "*"(A1: Matrix; A2: Tensor) return Tensor;
  function Mult(A1: Tensor; A2,A3: Matrix) return Tensor;
  procedure MultAdd(A1,A2: in Matrix; A3: in out Matrix);

--  procedure UNPACK(VA: in Vector; Alpha,Tau: in Scalar; VB: out Vector);

  subtype Polynom1 is Vector;  -- assuming index starting at 0
  subtype Polynom2 is Matrix;  -- assuming indices starting at 0

  procedure SetConst(S: in Scalar; P: out Polynom1);
  function Val(X: Scalar; P: Polynom1) return Scalar;
  procedure Scale(S: in Scalar; P: in out Polynom1);
  procedure Scale(S: in Scalar; P1: in Polynom1; P2: out Polynom1);
  procedure Translate(S: in Scalar; P1: in Polynom1; P2: out Polynom1);
  procedure SetConst(S: in Scalar; P: in out Polynom2);
  function Val(X,Y: Scalar; P: Polynom2) return Scalar;
  procedure Scale(S1,S2: in Scalar; P: in out Polynom2);
  procedure Scale(S1,S2: in Scalar; P1: in Polynom2; P2: out Polynom2);
  procedure Translate1(S: in Scalar; P1: in Polynom2; P2: out Polynom2);
  procedure Translate2(S: in Scalar; P1: in Polynom2; P2: out Polynom2);
  procedure Translate(S1,S2: in Scalar; P1: in Polynom2; P2: out Polynom2);

end Vectors;
